/* creates, parses, and prints allnet addresses. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include "types.h"

/* stores an address for sending on the net.
 * if successful, returns the number of bytes used.
 * If bsize is insufficient, returns - the size needed,
 * e.g. -8 for an IPv4 address.
 */
int store_addr (struct allnet_addr * addr, char * buffer, int bsize)
{
  int offset = 2;
  if (bsize >= offset)
    bs2 (buffer, addr->discrim);

  int i, bytes, extra;
  extra = 0;

  switch (addr->discrim) {
  case ALLNET_NO_ADDR_ID:
    break;
  case ALLNET_BCAST_ADDR_ID:
    if (bsize >= offset + 1)
      buffer [offset] = b1 (addr->bc_addr.hops);
    offset += 1;
    break;
  case ALLNET_UNICAST_ADDR_ID:
    if (bsize >= offset + 2) {
      buffer [offset] = b1 (addr->uc_addr.hops);
      buffer [offset + 1] = b1 (addr->uc_addr.num_bytes);
    }
    offset += 2;
    bytes = b1 (addr->uc_addr.num_bytes);
    if (bsize >= offset + bytes) {
      for (i = 0; i < bytes; i++) {
        buffer [offset] = b1 (addr->uc_addr.id + i);
        offset++;
      }
    }
    break;
  case ALLNET_KEY_ADDR_ID:
    if (bsize >= offset + 2) {
      buffer [offset] = b1 (addr->pk_addr.hops);
      buffer [offset + 1] = b1 (addr->pk_addr.num_bits);
    }
    offset += 2;
    bytes = (b1 (addr->pk_addr.num_bits) + 7) / 8;
    if (bsize >= offset + bytes) {
      for (i = 0; i < bytes; i++) {
        buffer [offset] = b1 (addr->pk_addr.pubkey + i);
        offset++;
      }
    }
    break;
  case ALLNET_IP4_ADDR_ID:
    if (bsize >= offset + 6) {
      memcpy (buffer + offset, &(addr->i4_addr.ip4), 4);
      memcpy (buffer + offset + 4, &(addr->i4_addr.port), 2);
    }
    offset += 6;
    break;
  case ALLNET_IP6_ADDR_ID:
    if (bsize >= offset + 22) {
      memcpy (buffer + offset, &(addr->i6_addr.ip6), 16);
      memcpy (buffer + offset + 16, &(addr->i6_addr.port), 2);
      memcpy (buffer + offset + 18, &(addr->i6_addr.scope), 4);
    }
    offset += 22;
    break;
  case ALLNET_GEO_ADDR_ID_LLRHR:
    extra = 4;
    bs4 (buffer + offset + 16, addr->geo_addr.height_range);
  case ALLNET_GEO_ADDR_ID_LLRH:
    extra += 4;
    bs4 (buffer + offset + 12, addr->geo_addr.height);
  case ALLNET_GEO_ADDR_ID_LLR:
    extra += 4;
    bs4 (buffer + offset + 8, addr->geo_addr.radius);
  case ALLNET_GEO_ADDR_ID_LL:
    bs4 (buffer + offset + 4, addr->geo_addr.center_lon);
    bs4 (buffer + offset    , addr->geo_addr.center_lat);
    offset += 8 + extra;
    break;
  default:
    printf ("error: unknown discriminator %d in store_addr\n", addr->discrim);
    exit (1);
  }
  if (offset <= bsize)
    return offset;
  else
    return -offset;
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the address is stored in *result;
 */
int parse_addr (char * buffer, int bsize, struct allnet_addr * result)
{
  if (bsize < 2) return -1;
  result->discrim = b2 (buffer);
  int geo_size = 0;
  switch (result->discrim) {
    case ALLNET_NO_ADDR_ID:
      return 2;    /* done */
    case ALLNET_BCAST_ADDR_ID:
      if (bsize < 3) return -1;
      result->bc_addr.hops [0] = b1 (buffer + 2);
      return 3;
    case ALLNET_KEY_ADDR_ID:
      if (bsize < 4) return -1;
      result->pk_addr.hops [0] = b1 (buffer + 2);
      result->pk_addr.num_bits [0] = b1 (buffer + 3);
      if (result->pk_addr.num_bits > 0) {
        int expected_bytes = (b1 (result->pk_addr.num_bits) + 7) / 8;
        if (expected_bytes > 16) return -1;
        if (bsize < 4 + expected_bytes) return -1;
        memcpy (result->pk_addr.pubkey, buffer + 4, expected_bytes);
        return 4 + expected_bytes;
      }
    case ALLNET_IP4_ADDR_ID:
      if (bsize < 8) return -1;
      memcpy (&(result->i4_addr.ip4), buffer + 2, 4);
      memcpy (&(result->i4_addr.port), buffer + 6, 2);
      return 8;
    case ALLNET_IP6_ADDR_ID:
      if (bsize < 24) return -1;
      memcpy (&(result->i6_addr.ip6), buffer + 2, 16);
      memcpy (&(result->i4_addr.port), buffer + 18, 2);
      result->i6_addr.scope = b4 (buffer + 20);
      return 24;
    case ALLNET_GEO_ADDR_ID_LLRHR:
      if (bsize < 22) return -1;
      geo_size = 22;
      result->geo_addr.height_range = b4 (buffer + 18);
    case ALLNET_GEO_ADDR_ID_LLRH:
      if (bsize < 18) return -1;
      if (geo_size == 0) geo_size = 18;
      result->geo_addr.height = b4 (buffer + 14);
    case ALLNET_GEO_ADDR_ID_LLR:
      if (bsize < 14) return -1;
      if (geo_size == 0) geo_size = 14;
      result->geo_addr.radius = b4 (buffer + 10);
    case ALLNET_GEO_ADDR_ID_LL:
      if (bsize < 10) return -1;
      if (geo_size == 0) geo_size = 10;
      result->geo_addr.center_lon = b4 (buffer + 6);
      result->geo_addr.center_lat = b4 (buffer + 2);
      return geo_size;
    default:
      printf ("unknown address type %d/%x\n", result->discrim, result->discrim);
      int * foo = NULL;
      *foo = 3;
      return -1;
  }
}

/* if the result of calling snprintf is < ssize, return that plus 1, otherwise
 * its negative
 */
#define ret_result() if (result < 0) return 0; \
                     return (result + 1) * ((result >= ssize) ? (-1) : 1);

/* makes the address printable, adding a '\0' at the end, returning the
 * number of bytes written (including the '\0').
 * If bsize is insufficient, returns - the size needed,
 * e.g. -21 for an IPv4 address such as 123.123.123.123/12345.
 * zero is returned in case of other errors.
 * in case of any error, the string may be garbled, do not use.
 */
int addr_to_string (struct allnet_addr * addr, char * string, int ssize)
{
  int result;
  if (addr == NULL) {
    result = snprintf (string, ssize, "(null addr)\n");
    ret_result ();
  }
/*
  printf ("address type %d (%x) for %p\n", addr->discrim, addr->discrim, addr);
*/
  int i, number, bound, needed, hops;
  /* for printing internet addresses.  Size is slightly too big, which is OK */
  static char ipaddr [INET6_ADDRSTRLEN + INET_ADDRSTRLEN + 1];
  double lat, lon, radius, height, hr;
  switch (addr->discrim) {
  case ALLNET_NO_ADDR_ID:
    result = snprintf (string, ssize, "no address\n");
    exit (1);
    ret_result ();
  case ALLNET_BCAST_ADDR_ID:
    result = snprintf (string, ssize, "broadcast address %d hops max\n",
                       b1 (addr->bc_addr.hops));
    ret_result ();
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_UNICAST_ADDR_ID:
    hops = b1 (addr->pk_addr.hops);
    number = b1 (addr->pk_addr.num_bits);
    if (addr->discrim == ALLNET_KEY_ADDR_ID)
      result = snprintf (string, ssize,
                         "public key address with %d hops, %d bits: ",
                         hops, number);
    else
      result = snprintf (string, ssize,
                         "unicast address with %d hops, %d bytes: ",
                         hops, number);
    if (result < 0) return 0;
    if (result + 1 < ssize) { string += result; ssize -= result; }
    needed = result;

    bound = ((number + 7) / 8);
    if (addr->discrim == ALLNET_UNICAST_ADDR_ID)
      bound = number;
    for (i = 0; i < bound; i++) {
      if (i + 1 < bound)
        result = snprintf (string, ssize, "0x%02x, ",
                           b1 (addr->pk_addr.pubkey + i));
      else
        result = snprintf (string, ssize, "0x%02x",
                           b1 (addr->pk_addr.pubkey + i));
      if (result < 0) return 0;
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    result = snprintf (string, ssize, "\n");
    if (result < 0) return 0;
    needed += result;
    result = needed;
    ret_result ();
    return;
  case ALLNET_IP4_ADDR_ID:
    if (inet_ntop (AF_INET, &(addr->i4_addr.ip4), ipaddr, sizeof (ipaddr))
        == NULL) {
      perror ("inet_ntop/4");
      printf ("unprintable ipv4, port %d/%x\n",
              ntohs (addr->i4_addr.port), ntohs (addr->i4_addr.port));
      return 0;
    }
    result = snprintf (string, ssize, "ipv4 address %s, port %d/%x",
                       ipaddr, ntohs (addr->i4_addr.port),
                       ntohs (addr->i4_addr.port));
    ret_result ();
  case ALLNET_IP6_ADDR_ID:
    if (inet_ntop (AF_INET6, &(addr->i6_addr.ip6), ipaddr, sizeof (ipaddr))
        == NULL) {
      perror ("inet_ntop/6");
      printf ("unprintable ipv6, port %d/%x\n", ntohs (addr->i6_addr.port),
              ntohs (addr->i6_addr.port));
      return 0;
    }
    result = snprintf (string, ssize, "ipv6 address %s, port %d/%x, scope %d",
                       ipaddr, ntohs (addr->i6_addr.port),
                       ntohs (addr->i6_addr.port), ntohs (addr->i6_addr.scope));
    ret_result ();
  case ALLNET_GEO_ADDR_ID_LL:
  case ALLNET_GEO_ADDR_ID_LLR:
  case ALLNET_GEO_ADDR_ID_LLRH:
  case ALLNET_GEO_ADDR_ID_LLRHR:
    lat = ((double) addr->geo_addr.center_lat) * 180.0;
    lon = ((double) addr->geo_addr.center_lon) * 180.0;
    result = snprintf (string, ssize, "lat %f, lon %f", lat, lon);
    if (result < 0) return 0;
    if (result + 1 < ssize) { string += result; ssize -= result; }
    needed = result;

    if ((addr->discrim == ALLNET_GEO_ADDR_ID_LLR) ||
        (addr->discrim == ALLNET_GEO_ADDR_ID_LLRH) ||
        (addr->discrim == ALLNET_GEO_ADDR_ID_LLRHR)) {
      result = snprintf (string, ssize, ", radius %f",
                         ((double) addr->geo_addr.radius) * 180.0);
      if (result < 0) return 0;
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    if ((addr->discrim == ALLNET_GEO_ADDR_ID_LLRH) ||
        (addr->discrim == ALLNET_GEO_ADDR_ID_LLRHR)) {
      result = snprintf (string, ssize, ", height %.2f",
                         ((double) addr->geo_addr.height) / 100.0);
      if (result < 0) return 0;
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    if (addr->discrim == ALLNET_GEO_ADDR_ID_LLRHR) {
      result = snprintf (string, ssize, " +- %.2f",
                         ((double) addr->geo_addr.height_range) / 100.0);
      if (result < 0) return 0;
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    result = printf (string, ssize, "\n");
    if (result < 0) return 0;
    needed += result;
    result = needed;
    ret_result ();
  default:
    printf (string, ssize, "unknown address type %d\n", addr->discrim);
    return 0;
  }
}

/* prints the address, optionally with a newline at the end (if eof > 0) */
void print_addr (struct allnet_addr * addr, int eol)
{
  int bufsize = 10000;
  char * buffer = malloc (bufsize);
  if (buffer != NULL) {
    int printed = addr_to_string (addr, buffer, bufsize);
    if (printed > 0) {
      if (eol > 0)
        printf ("%s\n", buffer);
      else
        printf ("%s", buffer);
    } else {
      printf ("error printing address\n");
    } 
    free (buffer);
  } else {
    printf ("memory error, unable to print address\n");
  }
}

/* returns 1 if the addresses are the same, 0 if they differ */
int same_addr (struct allnet_addr * a1, struct allnet_addr * a2)
{
  if (a1->discrim != a2->discrim)
    return 0;
  int i, n, bits, bytes;
  switch (a1->discrim) {
  case ALLNET_NO_ADDR_ID:
    return 1;                    /* same address */
  case ALLNET_BCAST_ADDR_ID:
    /* two broadcast addresses are always the same, even if the number of
     * hops is different. */
    return 1;                    /* same address */
  case ALLNET_UNICAST_ADDR_ID:
    /* again, ignore the number of hops */
    if (a1->uc_addr.num_bytes [0] != a2->uc_addr.num_bytes [0]) return 0;
    bytes = b1 (a1->uc_addr.num_bytes);
    if (memcmp (a1->uc_addr.id, a2->uc_addr.id, bytes) != 0) return 0;
    return 1;
  case ALLNET_KEY_ADDR_ID:
    /* the number of hops does not affect equality */
    if (a1->pk_addr.num_bits [0] != a2->pk_addr.num_bits [0]) return 0;
    n = (b1 (a1->pk_addr.num_bits) + 7) / 8;
    for (i = 0; i + 1 < n; i++)
      if (a1->pk_addr.pubkey [i] != a2->pk_addr.pubkey [i]) return 0;
    /* just compare the first few bits of the last byte */ 
    bits = b1 (a1->pk_addr.num_bits) % 8;
    printf ("bytes 0..%d are the same, last byte (%d bits) is %x =? %x\n",
            n - 2, bits,
            b1 (a1->pk_addr.pubkey + n - 1), b1 (a2->pk_addr.pubkey + n - 1));
    if (bits == 0) {
      if (a1->pk_addr.pubkey [n - 1] != a2->pk_addr.pubkey [n - 1]) return 0;
      printf ("last 8 bits are the same\n");
      return 1;
    }
    /* 8 > bits > 0 */
    if ((b1 (a1->pk_addr.pubkey + n - 1) >> (8 - bits)) !=
        (b1 (a2->pk_addr.pubkey + n - 1) >> (8 - bits)))
      return 0;
    printf ("last %d bits are the same\n", bits);
    return 1;                    /* same address */
  case ALLNET_IP4_ADDR_ID:
    if (a1->i4_addr.ip4.s_addr != a2->i4_addr.ip4.s_addr) return 0;
    if (a1->i4_addr.port != a2->i4_addr.port) return 0;  /* different */
    return 1;                    /* same address */
  case ALLNET_IP6_ADDR_ID:
    if ((a1->i6_addr.ip6.s6_addr32 [0] != a2->i6_addr.ip6.s6_addr32 [0]) ||
        (a1->i6_addr.ip6.s6_addr32 [1] != a2->i6_addr.ip6.s6_addr32 [1]) ||
        (a1->i6_addr.ip6.s6_addr32 [2] != a2->i6_addr.ip6.s6_addr32 [2]) ||
        (a1->i6_addr.ip6.s6_addr32 [3] != a2->i6_addr.ip6.s6_addr32 [3]))
      return 0;
    if (a1->i6_addr.port != a2->i6_addr.port) return 0;  /* different */
    /* not sure if we should compare scopes, but why not */
    if (a1->i6_addr.scope != a2->i6_addr.scope) return 0;  /* different */
    return 1;                    /* same address */
  case ALLNET_GEO_ADDR_ID_LLRHR:
    if (a1->geo_addr.height_range != a2->geo_addr.height_range) return 0;
  case ALLNET_GEO_ADDR_ID_LLRH:
    if (a1->geo_addr.height != a2->geo_addr.height) return 0;
  case ALLNET_GEO_ADDR_ID_LLR:
    if (a1->geo_addr.radius != a2->geo_addr.radius) return 0;
  case ALLNET_GEO_ADDR_ID_LL:
    if (a1->geo_addr.center_lat != a2->geo_addr.center_lat) return 0;
    if (a1->geo_addr.center_lon != a2->geo_addr.center_lon) return 0;
    return 1;                    /* same address */
  default:
    return 0;                   /* may be same address, but how do I know? */
  }
}

/* returns 1 if the address 1 matches address 2, 0 if it cannot. */
/* this is mostly the same as same_addr, but for keys and unicast
 * it allows a more generic a1 to match a more specific a2 */
int matches_addr (struct allnet_addr * a1, struct allnet_addr * a2)
{
  if (a1->discrim != a2->discrim)
    return 0;
  if (same_addr (a1, a2))
    return 1;
  /* only have to worry about keys, unicast, and geo addresses */
  int i, n, bits, bytes;
  switch (a1->discrim) {
  case ALLNET_UNICAST_ADDR_ID:
    if (a1->uc_addr.num_bytes [0] > a2->uc_addr.num_bytes [0]) return 0;
    bytes = b1 (a1->uc_addr.num_bytes);
    /* do the first bytes match? */
    if (memcmp (a1->uc_addr.id, a2->uc_addr.id, bytes) != 0) return 0;
    return 1;
  case ALLNET_KEY_ADDR_ID:
    /* again, allow a1 to have fewer bits than a2 */
    if (a1->pk_addr.num_bits [0] > a2->pk_addr.num_bits [0]) return 0;
    n = (b1 (a1->pk_addr.num_bits) + 7) / 8;
    for (i = 0; i + 1 < n; i++)
      if (a1->pk_addr.pubkey [i] != a2->pk_addr.pubkey [i]) return 0;
    /* just compare the first few bits of the last byte */ 
    bits = b1 (a1->pk_addr.num_bits) % 8;
    printf ("bytes 0..%d match, last byte (%d bits) is %x =? %x\n",
            n - 2, bits,
            b1 (a1->pk_addr.pubkey + n - 1), b1 (a2->pk_addr.pubkey + n - 1));
    if (bits == 0) {
      if (a1->pk_addr.pubkey [n - 1] != a2->pk_addr.pubkey [n - 1]) return 0;
      printf ("last 8 bits match\n");
      return 1;
    }
    /* 8 > bits > 0 */
    if ((b1 (a1->pk_addr.pubkey + n - 1) >> (8 - bits)) !=
        (b1 (a2->pk_addr.pubkey + n - 1) >> (8 - bits)))
      return 0;
    printf ("last %d bits match\n", bits);
    return 1;                    /* same address */
    if (a1->pk_addr.num_bits [0] > a2->pk_addr.num_bits [0]) return 0;
    n = (b1 (a1->pk_addr.num_bits) + 7) / 8;
    for (i = 0; i + 1 < n; i++)
      if (a1->pk_addr.pubkey [i] != a2->pk_addr.pubkey [i]) return 0;
    /* just compare the first few bits of the last byte */ 
    bits = b1 (a1->pk_addr.num_bits) % 8;
    printf ("bytes 0..%d match, last byte is %x =? %x\n",
            n - 2, b1 (a1->pk_addr.pubkey + n - 1) >> (8 - bits),
                   b1 (a2->pk_addr.pubkey + n - 1) >> (8 - bits));
    if (bits == 0) {
      if (a1->pk_addr.pubkey [n - 1] != a2->pk_addr.pubkey [n - 1]) return 0;
      printf ("last 8 bits are the same\n");
      return 1;
    }
    /* 8 > bits > 0 */
    if ((b1 (a1->pk_addr.pubkey + n - 1) >> (8 - bits)) !=
        (b1 (a2->pk_addr.pubkey + n - 1) >> (8 - bits)))
      return 0;
    printf ("last %d bits are the same\n", bits);
    return 1;                    /* same address */
  case ALLNET_GEO_ADDR_ID_LLRHR:
    /* is the center of this address within the range of the other address? */
    printf ("geo address matching not implemented yet\n");  return 0;
    if (a1->geo_addr.height_range != a2->geo_addr.height_range) return 0;
  case ALLNET_GEO_ADDR_ID_LLRH:
    printf ("geo address matching not implemented yet\n");  return 0;
    if (a1->geo_addr.height != a2->geo_addr.height) return 0;
  case ALLNET_GEO_ADDR_ID_LLR:
    printf ("geo address matching not implemented yet\n");  return 0;
    if (a1->geo_addr.radius != a2->geo_addr.radius) return 0;
  case ALLNET_GEO_ADDR_ID_LL:
    printf ("geo address matching not implemented yet\n");  return 0;
    if (a1->geo_addr.center_lat != a2->geo_addr.center_lat) return 0;
    if (a1->geo_addr.center_lon != a2->geo_addr.center_lon) return 0;
    return 1;                    /* same address */
  case ALLNET_IP4_ADDR_ID:
  case ALLNET_IP6_ADDR_ID:
  /* should have been handled by same_addr, so they do not match */
    return 0;
  default:
    return 0;                   /* may match, but how do I know? */
  }
}
