/* allnet daemon, runs independently and forwards messages based on priority.
 *
 * this daemon occasionally listens and sends on wireless (protocol
 * number 0xA119) and constantly listens and occasionally sends allnet
 * traffic on UDP port 0xA119.
 *
 * this daemon also communicates with any local allnet clients on TCP
 * port 0xA11E.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <netpacket/packet.h>
#include <time.h>
#include <pthread.h>
#include <ifaddrs.h>

#include "types.h"
#include "messages.h"
#include "synqueue.h"
#include "peer_info.h"
#include "bloom.h"

#define MAX_RECEIVE_BUFFER	65507  /* max UDP size */
/* there is a send queue for every physical interface capable of sending
 * packets. */
struct send_queue_info {
  /* the queue itself */
  struct syn_queue * queue;
  char * id;               /* string meaningful to the user */
  int interface_index;     /* for sending on physical interfaces */
  int is_local;            /* is this a local application */
  int num_local_clients;
#define MAX_LOCAL_CLIENTS	100
  unsigned short local_ports [MAX_LOCAL_CLIENTS];
  /* normally either the physical or the internet sockets are set, not both */
  int physical_socket;     /* physical interface, -1 if not connected */
  int ip4_udp_socket;      /* -1 if not available */
  int ip6_udp_socket;      /* -1 if not available */
  int ip4_tcp_socket;      /* accepting (passive) socket, or -1 */
  int ip6_tcp_socket;      /* accepting (passive) socket, or -1 */
  /* one or more addresses of the local interface */
  int num_addrs;
  struct allnet_addr * my_addrs;
  /* inet/6 sockets have a local address */
  int has_inet_addr;
  struct sockaddr_storage my_inet_addr;
  /* record a speed limit for this interface here, otherwise 0 */
  int max_bytes_per_sec;
  /* not sure if we need these thread vars, but this is the place for them */
  pthread_t listen_thread;
  pthread_t send_thread;
};

#define MAX_SEND_QUEUES	20    /* probably too many, but better than too few */
struct send_queue_info send_queues [MAX_SEND_QUEUES];
static int num_send_queues = -1;

static void init_send_queues ()
{
  if (num_send_queues == -1) {
    num_send_queues = 0;
    int i;
    for (i = 0; i < MAX_SEND_QUEUES; i++) {
      send_queues [i].queue = NULL;
      send_queues [i].id = "";
      send_queues [i].interface_index = -1;
      send_queues [i].is_local = 0;
      send_queues [i].num_local_clients = 0;
      send_queues [i].physical_socket = -1;
      send_queues [i].ip4_udp_socket = -1;
      send_queues [i].ip6_udp_socket = -1;
      send_queues [i].ip4_tcp_socket = -1;
      send_queues [i].ip6_tcp_socket = -1;
      send_queues [i].num_addrs = 0;
      send_queues [i].my_addrs = NULL;
      send_queues [i].has_inet_addr = 0;
      send_queues [i].max_bytes_per_sec = 0;  /* unspecified */
    }
  }
}

#define MAX_LOCAL_INET_ADDRS	100
static struct sockaddr_storage my_inet_addrs [MAX_LOCAL_INET_ADDRS];
static int num_my_inet_addrs = 0;

static char * copy_string (char * string)
{
  char * result = malloc (strlen (string) + 1);
  strcpy (result, string);
  return result;
}

static void copy_addr (struct sockaddr_storage * dest, struct sockaddr * src)
{
  struct sockaddr_in  * sin  = (struct sockaddr_in  *) src;
  struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) src;
  struct sockaddr_ll  * sll  = (struct sockaddr_ll  *) src;
  if (sin->sin_family == AF_INET) {
    struct sockaddr_in  * d  = (struct sockaddr_in  *) dest;
    *d = *sin;
  } else if (sin6->sin6_family == AF_INET6) {
    struct sockaddr_in6 * d  = (struct sockaddr_in6 *) dest;
    *d = *sin6;
  } else if (sin6->sin6_family == AF_PACKET) {
    struct sockaddr_ll  * d  = (struct sockaddr_ll  *) dest;
    *d = *sll;
  }
}

void * listen_loop (void * arg);
void * send_loop (void * arg);

static int new_send_queue (char * name, int ifn, int max_size, int is_local,
                           int max_bytes_per_sec, int phys_socket,
                           int ip4_udp, int ip6_udp, int ip4_tcp, int ip6_tcp,
                           int num_addrs, struct allnet_addr * addrs,
                           int has_inet_addr, struct sockaddr * inet_addr)
{
  if (num_send_queues >= MAX_SEND_QUEUES) {
    printf ("too many send interfaces, too many send queues \n");
    exit (1);
  }
  struct send_queue_info * q = send_queues + num_send_queues;
  int index = num_send_queues;
  num_send_queues++;
  q->queue = new_syn_queue (max_size);
  q->id = copy_string (name);
  q->interface_index = ifn;
  q->is_local = is_local;
  q->physical_socket = phys_socket;
  q->ip4_udp_socket = ip4_udp;
  q->ip6_udp_socket = ip6_udp;
  q->ip4_tcp_socket = ip4_tcp;
  q->ip6_tcp_socket = ip6_tcp;
  q->num_addrs = num_addrs;
  if (num_addrs > 0) {
    int size = sizeof (struct allnet_addr) * num_addrs;
    q->my_addrs = malloc (size);
    /* if result is NULL, segfault */
    memcpy (q->my_addrs, addrs, size);
  }
  q->has_inet_addr = has_inet_addr;
  if (has_inet_addr)
    copy_addr (&(q->my_inet_addr), inet_addr);
  q->max_bytes_per_sec = max_bytes_per_sec;
  if ((q->physical_socket != -1) ||   /* if at least one socket is valid */
      (q->ip4_udp_socket != -1) || (q->ip6_udp_socket != -1) ||
      (q->ip4_tcp_socket != -1) || (q->ip6_tcp_socket != -1)) {
    void * arg1 = malloc (sizeof (int));
    void * arg2 = malloc (sizeof (int));
         /* again, segfault if malloc fails */
    * ((int *) arg1) = index;
    * ((int *) arg2) = index;
    printf ("initializing threads for interface %d\n", index);
    pthread_create (&(q->listen_thread), NULL, listen_loop, arg1);
    pthread_create (&(q->send_thread), NULL, send_loop, arg2);
  }
  return index;
} 

static void inet_name (struct sockaddr * addr, char * name, int nlen)
{
  struct sockaddr_in  * sin  = (struct sockaddr_in  *) addr;
  struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) addr;
  struct sockaddr_ll  * sll  = (struct sockaddr_ll  *) addr;

  int port = -1;
  if (sin->sin_family == AF_INET) {
    snprintf (name, nlen, "IP4: ");
    inet_ntop (sin->sin_family, &(sin->sin_addr), name + 5, nlen - 5);
    port = ntohs (sin->sin_port);
  } else if ((sin6->sin6_family == AF_INET6) &&
             (sin6->sin6_addr.s6_addr32 [0] == 0) &&
             (sin6->sin6_addr.s6_addr32 [1] == 0) &&
             (ntohl (sin6->sin6_addr.s6_addr32 [2]) == 0xffff)) {
    /* IPv6 encoded IPv4 address */
    snprintf (name, nlen, "IP4: ");
    struct in_addr ina;
    ina.s_addr = sin6->sin6_addr.s6_addr32 [3];
    port = ntohs (sin6->sin6_port);
    inet_ntop (AF_INET, &ina, name + 5, nlen - 5);
  } else if (sin6->sin6_family == AF_INET6) {
    snprintf (name, nlen, "IP6: ");
    inet_ntop (sin6->sin6_family, &(sin6->sin6_addr), name + 5, nlen - 5);
    port = ntohs (sin6->sin6_port);
  } else if (sin6->sin6_family == AF_PACKET) {
    int c = snprintf (name, nlen, "if %d proto %d hlen %d addr %02x",
                      sll->sll_ifindex, sll->sll_protocol, sll->sll_halen,
                      b1 (sll->sll_addr));
    int i;
    for (i = 1; i < sll->sll_halen; i++) {
      if (c < nlen)
        c += snprintf (name + c, nlen - c, ":%02x", b1 (sll->sll_addr + i));
    }
  } else {
    printf ("unknown address family %d in inet_name\n", sin->sin_family);
    snprintf (name, nlen, "unknown");
  }
  if (port != -1) {
    int len = strlen (name);
    if (len < nlen)
      snprintf (name + len, nlen - len, " port %d/%x", port, port);
  }
}

static void print_send_queues ()
{
  int i;
  char a [1000];
  for (i = 0; i < num_send_queues; i++) {
    struct send_queue_info * q = send_queues + i;
    printf ("sq [%d] = %s%s, ", i, q->id, (q->is_local) ? " (local)" : "");
    if (q->physical_socket != -1)
      printf ("physical socket %d, ", q->physical_socket);
    if (q->ip4_udp_socket != -1)
      printf ("ip4 udp socket %d, ", q->ip4_udp_socket);
    if (q->ip6_udp_socket != -1)
      printf ("ip6 udp socket %d, ", q->ip6_udp_socket);
    if (q->ip4_tcp_socket != -1)
      printf ("ip4 tcp socket %d, ", q->ip4_tcp_socket);
    if (q->ip6_tcp_socket != -1)
      printf ("ip6 tcp socket %d, ", q->ip6_tcp_socket);
    if (q->num_addrs > 0) {
      if (q->num_addrs == 1)
        printf ("1 allnet address: ", q->num_addrs);
      else
        printf ("%d allnet addresses: ", q->num_addrs);
      int j;
      for (j = 0; j < q->num_addrs; q++) {
        print_addr (q->my_addrs + j, 0);
        printf (", ");
      }
    }
    if (q->has_inet_addr) {
      inet_name (((struct sockaddr *) &(q->my_inet_addr)), a, sizeof (a));
      printf ("internet address %s ", a);
    }
    printf ("\n");
  }
}

static void print_buffer (char * buffer, int size, int print_eol)
{
  while (size-- > 0) {
    printf ("%02x, ", (*buffer) & 0xff);
    buffer++;
  }
  if (size >= 0)
    printf ("%02x", (*buffer) & 0xff);
  if (print_eol)
    printf ("\n");
}

/* returns 1 if this is a local (my) address, 0 otherwise */
static int is_my_addr (struct allnet_addr * a)
{
  int i, j;
  for (i = 0; i < num_send_queues; i++) {
    for (j = 0; j < send_queues [i].num_addrs; j++) {
/*
      printf ("comparison of ");
      print_addr (a, 0);
      printf ("to ");
      print_addr (&(send_queues [i].my_addrs [j]), 1);
*/
      if (matches_addr (a, &(send_queues [i].my_addrs [j])))
        return 1;
    }
  }  
  for (i = 0; i < num_my_inet_addrs; i++) {
    struct sockaddr * sap = (struct sockaddr *) (& (my_inet_addrs [i]));
/*
    char string [100];
    inet_name (sap, string, sizeof (string));
    printf ("comparing %s to ", string);
    print_addr (a, 1);
*/
    if ((a->discrim == ALLNET_IP4_ADDR_ID) && (sap->sa_family == AF_INET)) {
      struct sockaddr_in  * sin  = (struct sockaddr_in  *) sap;
      if ((sin->sin_addr.s_addr == a->i4_addr.ip4.s_addr) &&
          ((sin->sin_port == a->i4_addr.port) || (sin->sin_port == 0)))
        return 1;
    }
    if ((a->discrim == ALLNET_IP6_ADDR_ID) && (sap->sa_family == AF_INET6)) {
      struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) sap;
      if ((sin6->sin6_addr.s6_addr32 [0] == a->i6_addr.ip6.s6_addr32 [0]) &&
          (sin6->sin6_addr.s6_addr32 [1] == a->i6_addr.ip6.s6_addr32 [1]) &&
          (sin6->sin6_addr.s6_addr32 [2] == a->i6_addr.ip6.s6_addr32 [2]) &&
          (sin6->sin6_addr.s6_addr32 [3] == a->i6_addr.ip6.s6_addr32 [3]) &&
          ((sin6->sin6_port == a->i6_addr.port) || (sin6->sin6_port == 0)))
        return 1;
    }
  }
  return 0;
}

static int compute_priority (char * data, allnet_data_len data_size,
                             struct sockaddr * from, int is_local)
{
  if (is_local)
    return MAX_PRIORITY - 1;
  return MAX_PRIORITY / 3;
}

/* returns 1 if this packet is for this interface, 0 otherwise */
static int for_local_interface (struct allnet_route * route, int interface)
{
  if (! (send_queues [interface].is_local))
    return 0;
  int i;
  if (b1 (route->used) <= b1 (route->num_addrs)) {
    struct allnet_addr * used_addr = route->addrs + (b1 (route->used));
    for (i = 0; i < send_queues [interface].num_addrs; i++)
      if (matches_addr (used_addr, send_queues [interface].my_addrs + i))
        return 1;
  }
  if (b1 (route->used) < b1 (route->num_addrs)) {
    struct allnet_addr * next_addr = route->addrs + (b1 (route->used)) + 1;
    for (i = 0; i < send_queues [interface].num_addrs; i++)
      if (matches_addr (next_addr, send_queues [interface].my_addrs + i))
        return 1;
  }
  return 0;
}

/* returns 1 if the packet can be forwarded to non-local destinations,
 * returns 0 if the packet has reached the end of its route and
 * should not be forwarded, and
 * returns -1 if the packet was seen before */
static int forwarding_decision (struct allnet_route * route,
                                char * data, allnet_data_len data_len)
{
  if (bloom_match (data, data_len)) { /* found, do nothing */
    printf ("packet matches earlier packet in bloom filter\n");
    return -1;
  }
  /* not in the bloom filters, will forward, so add it to the filters */
  bloom_add (data, data_len);
  /* decide whether we are at the end of the route and, as needed,
   * update the "used" field */
  /* in the future, should probably also update broadcast routes so they
   * record the return route */
  if (b1 (route->used) > b1 (route->num_addrs))
    return 0;   /* used more than available, only forward locally */
  struct allnet_addr * a = route->addrs + b1 (route->used);
  int i, j;
  int matched = 0;
  switch (a->discrim) {
  case ALLNET_UNICAST_ADDR_ID:
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_IP4_ADDR_ID:
  case ALLNET_IP6_ADDR_ID:
    for (i = 0; (i < num_send_queues) && (! matched); i++) {
      for (j = 0; (j < send_queues [i].num_addrs) && (! matched); j++) {
        /* if it is an exact match, only forward if there are more addresses */
        if ((same_addr (a, send_queues [i].my_addrs + j)) &&
            (route->used == route->num_addrs))
          return 0;   /* we are it, only forward locally */
        if (matches_addr (a, send_queues [i].my_addrs + j)) {
          route->used [0]++;       /* use the next address, if any */
          matched = 1;         /* we are this hop's destination */
        }
      }
    }
    if (matched)               /* already updated the "used" field */
      break;
    if ((a->discrim == ALLNET_IP4_ADDR_ID) ||
        (a->discrim == ALLNET_IP6_ADDR_ID)) {
      route->used [0]++;         /* use the next address, if any */
      break;
    }  /* else, do as if we got a broadcast */
  case ALLNET_BCAST_ADDR_ID:
    if (b1 (a->all_hops) > 1) {
      a->all_hops [0]--;   /* decrement hop count */
      return 1;            /* forward everywhere */
    }
    route->used [0]++;       /* use the next address, if any */
    break;
  default:
    printf ("not forwarding packets with destination of type %d\n",
            a->discrim);
    return 0;
  }
  if (b1 (route->used) > b1 (route->num_addrs))
    return 0;   /* no more addresses, only forward locally */
  return 1;     /* go ahead and forward */
}

/* if the buffer determines a single destination address, places that address
 * into *sap, and returns 1.
 *
 * if this is a broadcast message sent to an IP network, randomly
 * selects up to max of the peers, places their addresses into *sap,
 * and returns the total number of peers (which may be 1 up to max)
 *
 * otherwise returns 0 and *sap should not be used.
 *
 * it is assumed that sap refers to an array of max or more
 * struct sockaddr_storage
 */
static int addresses_from_buffer (char * buffer, int buffer_size,
                                  struct sockaddr_storage * sap, int max,
                                  int interface)
{
  struct sockaddr_ll * sll;

  struct allnet_header h;
  h.route.addrs = NULL;
  int hsize = parse_header (buffer, buffer_size, &h);

  if ((b1 (h.route.num_addrs) <= 0) ||
      ((b1 (h.route.num_addrs) > 0) &&
       (b1 (h.route.used) > b1 (h.route.num_addrs)))) {
    free (h.route.addrs);
    return 0;
  }

  struct allnet_addr next = h.route.addrs [b1 (h.route.used)];
  free (h.route.addrs);
  printf ("addresses_from_buffer: ");
  print_addr (&next, 1);

  if (((next.discrim == ALLNET_IP4_ADDR_ID) ||
       (next.discrim == ALLNET_IP6_ADDR_ID)) &&
      (((send_queues [interface].is_local) && (! is_my_addr (&next))) ||
       ((! send_queues [interface].is_local) && (is_my_addr (&next))))) {
    printf ("addresses_from: wrong kind of destination address for %s queue\n",
            (send_queues [interface].is_local) ? "loopback" : "non-local");
    return 0;
  }

  int i, result;
  char string [1000];  /* debugging */
  struct peer_info random [SENT_PEERS];
  if (max > SENT_PEERS)
    max = SENT_PEERS;

  printf ("addresses_from_buffer, address type %d\n", next.discrim);
  switch (next.discrim) {
  case ALLNET_NO_ADDR_ID:
    return 0;
  case ALLNET_BCAST_ADDR_ID:
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_UNICAST_ADDR_ID:
    if (send_queues [interface].interface_index != -1) {
      sll = (struct sockaddr_ll *) sap;
      sll->sll_family = AF_PACKET;
      sll->sll_protocol = ALLNET_WIFI_PROTOCOL;
      sll->sll_ifindex = send_queues [interface].interface_index;
      sll->sll_hatype = 0;  /* not used */
      sll->sll_pkttype = 0;  /* not used */
      sll->sll_halen = 6;
      sll->sll_addr [0] = sll->sll_addr [1] = sll->sll_addr [2] = 
        sll->sll_addr [3] = sll->sll_addr [4] = sll->sll_addr [5] = 0xff;
      if ((next.discrim == ALLNET_UNICAST_ADDR_ID) &&
          (b1 (next.uc_addr.num_bytes) >= 6)) {
        sll->sll_addr [0] = b1 (next.uc_addr.id + 0);
        sll->sll_addr [1] = b1 (next.uc_addr.id + 1);
        sll->sll_addr [2] = b1 (next.uc_addr.id + 2);
        sll->sll_addr [3] = b1 (next.uc_addr.id + 3);
        sll->sll_addr [4] = b1 (next.uc_addr.id + 4);
        sll->sll_addr [5] = b1 (next.uc_addr.id + 5);
      }
      inet_name ((struct sockaddr *) sll, string, sizeof (string));
      printf ("interface %i, using address %s\n", interface, string);
      return 1;
    }
    if ((send_queues [interface].ip4_udp_socket == -1) &&
        (send_queues [interface].ip6_udp_socket == -1)) {
      printf ("interface %i, neither physical nor inet\n", interface);
      return 0;
    }
    /* else, fall through into the next case -- multicasting to some peers */
  case ALLNET_IP4_ADDR_ID:
  case ALLNET_IP6_ADDR_ID:
    result = random_peers (random, max, 1, 0);
    if (result < 1) {
      printf ("addresses_from_buffer, no peers\n");
      return 0;   /* no internet peers */
    }
    for (i = 0; i < result; i++) {
      if (random [i].peer.peer.discrim == ALLNET_IP4_ADDR_ID) {
        struct sockaddr_in  * sin  = (struct sockaddr_in  *) (sap + i);
        sin->sin_family = AF_INET;
        sin->sin_addr = random [i].peer.peer.i4_addr.ip4;
        sin->sin_port = random [i].peer.peer.i4_addr.port;
      }
      if (random [i].peer.peer.discrim == ALLNET_IP6_ADDR_ID) {
        struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) (sap + i);
        sin6->sin6_family = AF_INET6;
        sin6->sin6_addr = random [i].peer.peer.i6_addr.ip6;
        sin6->sin6_port = random [i].peer.peer.i6_addr.port;
        sin6->sin6_scope_id = random [i].peer.peer.i6_addr.scope;
      }
    }
    return result;
  default:
    printf ("address type %d not handled in address_from_buffer\n",
            next.discrim);
    return 0;
  }
}

static void forward_packet (struct allnet_route * route, char * data,
                            allnet_data_len data_len, int priority, int clear)
{
  int decision = forwarding_decision (route, data, data_len);
  if (decision == -1) {
    printf ("packet seen before, discarding\n");
    return;
  }
  int i;
  for (i = 0; i < num_send_queues; i++) {
    if (((! (send_queues [i].is_local)) && (decision > 0)) ||
        ((send_queues [i].is_local) && (decision >= 0))) {
      int total_size = header_size (route) + data_len;
      /* need a new buffer for each entry in the queue */
      char * new_buffer = malloc (total_size);
      if (new_buffer == NULL) {
        printf ("error allocating new_buffer of size %d\n", total_size);
        return;
      }
      int r = store_data (data, data_len, clear, route, new_buffer, total_size);
      printf ("stored data message interface %i total size %d\n", i, total_size);
      if (r != total_size) {
        printf ("storing data message interface %i total size %d gave %d\n",
                i, total_size, r);
        free (new_buffer);
        return;
      }
      queue_add (send_queues [i].queue, new_buffer, total_size, priority);
    }
  }
}

/* returns 1 in case of success, 0 in case of failure */
static int new_hello_message (struct allnet_addr * destination,
                              struct allnet_peer * sender,
                              struct allnet_route * route,
                              char ** buffer, int * size)
{
  struct peer_info peers [SENT_PEERS];
  int num_peers = random_peers (peers, SENT_PEERS, 0, 0);
  int old_peers = num_peers;
  struct allnet_peer sending_peers [SENT_PEERS];
  int i = 0;
  while (i < num_peers) {
    if (same_addr (destination, &(peers [i].peer.peer))) {
      num_peers--;
    } else {
      sending_peers [i] = peers [i].peer;
      i++;
    }
  }
  printf ("new hello message, sending %d peers (was %d)\n",
          num_peers, old_peers);

  int bytes = sizeof (struct allnet_hello);
  bytes += num_peers * sizeof (struct allnet_peer);
  bytes += header_size (route);
  char * message = malloc (bytes);
  if (message == NULL)
    return 0;

  struct allnet_hello * hello;
  hello = malloc (sizeof (struct allnet_hello) +
                  sizeof (struct allnet_peer) * num_peers);
  if (hello == NULL) {
    printf ("unable to allocate hello message for sending\n");
    free (message);
    return 0;
  }

  int stored = store_hello (sender, num_peers, sending_peers, route,
                            message, bytes);
  free (hello);

  if (stored < 0)
    printf ("new_hello_message needed %d bytes, allocated %d bytes\n",
            -stored, bytes);
  if (stored <= 0) {
    free (message);
    printf ("error in new_hello_message, %d\n", stored);
    return 0;
  }
  if (stored < bytes) {
    message = realloc (message, stored);
    if (message == NULL) {
      printf ("error in new_hello_message: shrinking realloc failed\n");
      exit (1);
    }
    printf ("resized hello buffer from %d to %d bytes\n", bytes, stored);
    bytes = stored;
  }

  *buffer = message;
  *size = bytes;
  return 1;
}

/* returns 1 on success, 0 on failure */
static int get_addr4 (int interface, struct in_addr * ip, short * port)
{
  if (send_queues [interface].has_inet_addr) {
    struct sockaddr_in * sin = 
        (struct sockaddr_in *) &(send_queues [interface].my_inet_addr);
    if (sin->sin_family == AF_INET) {
      *ip = sin->sin_addr;
      *port = sin->sin_port;
      return 1;
    }
  }
  /* not necessarily an error
  printf ("unable to find IPv4 address for interface %d\n", interface); */
  return 0;
}

/* returns 1 on success, 0 on failure */
static int get_addr6 (int interface, struct in6_addr * ip, short * port,
                      unsigned int * scope)
{
  if (send_queues [interface].has_inet_addr) {
    struct sockaddr_in6 * sin = 
        (struct sockaddr_in6 *) &(send_queues [interface].my_inet_addr);
    if (sin->sin6_family == AF_INET6) {
      *ip = sin->sin6_addr;
      *port = sin->sin6_port;
      *scope = htonl (sin->sin6_scope_id);  /* I think scope is in host byte order */
      return 1;
    }
  }
  printf ("unable to find IPv6 address for interface %d\n", interface);
  return 0;
}

static void create_hello (struct allnet_addr * destination, int interface)
{
  struct allnet_route route;
  bs1 (route.num_addrs, 2);
  bs1 (route.used, 1);
  route.addrs = malloc (sizeof (struct allnet_addr) * 2);
  if (route.addrs == NULL) {
    printf ("unable to allocate ip4 route for hello message\n");
    return;
  }
  route.addrs [1] = *destination;

  /* set route.addrs [0] appropriately to my own address */
  int i, done;
  switch (destination->discrim) {
  case ALLNET_IP4_ADDR_ID:
  case ALLNET_IP6_ADDR_ID:
    route.addrs [0].discrim = destination->discrim;
    if (destination->discrim == ALLNET_IP4_ADDR_ID) {
      if (! get_addr4 (interface, &(route.addrs [0].i4_addr.ip4),
                       &(route.addrs [0].i4_addr.port))) {
        /* look for an IP 6 interface, should work */
        route.addrs [0].discrim = ALLNET_IP6_ADDR_ID;
        if (! get_addr6 (interface, &(route.addrs [0].i6_addr.ip6),
                         &(route.addrs [0].i6_addr.port),
                         &(route.addrs [0].i6_addr.scope)))
          /* no IP6 either */
          return;
      }
    } else { /* IP 6 */
      if (! get_addr6 (interface, &(route.addrs [0].i6_addr.ip6),
                       &(route.addrs [0].i6_addr.port),
                       &(route.addrs [0].i6_addr.scope)))
        return;
    }
    route.addrs [1] = *destination;
    break;
  case ALLNET_UNICAST_ADDR_ID:
  case ALLNET_BCAST_ADDR_ID:
  case ALLNET_KEY_ADDR_ID:
    done = 0;
    /* for my source address, use a key if available, otherwise a unicast
     * address, otherwise a broadcast address, in that order */
    for (i = 0; i < send_queues [interface].num_addrs; i++) {
      if (send_queues [interface].my_addrs [i].discrim == ALLNET_KEY_ADDR_ID) {
        route.addrs [0] = send_queues [interface].my_addrs [i];
        done = 1;
        break;
      }
    }
    if (done) break;
    for (i = 0; i < send_queues [interface].num_addrs; i++) {
      if (send_queues [interface].my_addrs [i].discrim
           == ALLNET_UNICAST_ADDR_ID) {
        route.addrs [0] = send_queues [interface].my_addrs [i];
        done = 1;
        break;
      }
    }
    if (done) break;
    for (i = 0; i < send_queues [interface].num_addrs; i++) {
      if (send_queues [interface].my_addrs [i].discrim == ALLNET_KEY_ADDR_ID) {
        route.addrs [0] = send_queues [interface].my_addrs [i];
        done = 1;
        break;
      }
    }
    break;
  case ALLNET_GEO_ADDR_ID_LL:
  case ALLNET_GEO_ADDR_ID_LLR:
  case ALLNET_GEO_ADDR_ID_LLRH:
  case ALLNET_GEO_ADDR_ID_LLRHR:
  case ALLNET_NO_ADDR_ID:  /* cannot send */
    printf ("create_hello, address type %d not supported, not creating\n",
            destination->discrim);
    free (route.addrs);
    return;
  default:
    printf ("create_hello, unknown address type %d\n", destination->discrim);
    free (route.addrs);
    return;
  }
  printf ("route address [0] = ");
  print_addr (route.addrs, 1);
  printf ("route address [1] = ");
  print_addr (route.addrs + 1, 1);

  struct allnet_peer myself;
  myself.provides = 0;
  myself.peer = route.addrs [0];

  char * buffer;
  int size;
  int made = new_hello_message (destination, &myself, &route, &buffer, &size);
  free (route.addrs);

  if (made) {
    printf ("adding message of size %d to queue %d\n", size, interface);
    print_buffer (buffer, size, 1);
    int result = queue_add (send_queues [interface].queue, buffer,
                            size, HELLO_PRIORITY);
    printf ("added message to queue %d\n", interface);
    if (result != size) {
      printf ("tried to enqueue %d bytes of hello ", size);
      printf ("on interface %d/%s ", interface, send_queues [interface].id);
      printf (", result was %d", result);
    }
  } else {
    printf ("unable to create hello message for interface %d\n", interface);
  }
}

void send_hello (struct peer_info * p)
{
  struct allnet_addr destination = p->peer.peer;
  int interface = -1;;
  int i;
  for (i = 0; i < num_send_queues; i++) {
    if ((destination.discrim == ALLNET_IP4_ADDR_ID) &&
        ((send_queues [i].ip4_udp_socket != -1) ||
         (send_queues [i].ip6_udp_socket != -1)) &&
        (send_queues [i].has_inet_addr)) {  /* matches, use this */
      interface = i;
      break;
    }
    if ((destination.discrim == ALLNET_IP6_ADDR_ID) &&
        (send_queues [i].ip6_udp_socket != -1) &&
        (send_queues [i].has_inet_addr)) {  /* matches, use this */
      interface = i;
      break;
    }
    if (((destination.discrim == ALLNET_BCAST_ADDR_ID) ||
         (destination.discrim == ALLNET_UNICAST_ADDR_ID) ||
         (destination.discrim == ALLNET_KEY_ADDR_ID)) &&
        (send_queues [i].num_addrs > 0)) {  /* matches, use this */
      interface = i;
      break;
    }
  }

  if (interface != -1) {
    printf ("sending hello to destination ");
    print_addr (&destination, 1);
    create_hello (&destination, interface);
    p->last_sent = time (NULL);
  } else {
    printf ("unable to find interface for peer ");
    print_addr (&destination, 1);
  }
}

static void send_hellos ()
{
  struct peer_info eligible [SAVED_PEERS];
  int num_peers = random_peers (eligible, SAVED_PEERS, 0, HELLO_INTERVAL);

  printf ("send_hellos called, sending to %d peers\n", num_peers);

  int i;
  for (i = 0; i < num_peers; i++)
    send_hello (eligible + i);
}

static void save_hello (struct allnet_hello * hello,
                        struct sockaddr * from,
                        struct allnet_route * source)
{
  if (! (is_my_addr (&(hello->sender.peer)))) {  /* not my own hello */
    printf ("adding sender: ");
    print_addr (&(hello->sender.peer), 1);
    struct allnet_addr from_addr;
    int add_from = 0;
    struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) from;
    if ((from->sa_family == AF_INET) ||    /* real IPv4 address */
        ((from->sa_family == AF_INET6) &&  /* IPv4 encoded as IPv6 */
         (sin6->sin6_addr.s6_addr32 [0] == 0) &&
         (sin6->sin6_addr.s6_addr32 [1] == 0) &&
         (ntohl (sin6->sin6_addr.s6_addr32 [2]) == 0xffff))) {
      from_addr.discrim = ALLNET_IP4_ADDR_ID;
      if (from->sa_family == AF_INET) {
        struct sockaddr_in * sin = (struct sockaddr_in *) from;
        from_addr.i4_addr.ip4 = sin->sin_addr;
        from_addr.i4_addr.port = sin->sin_port;
      } else {
        from_addr.i4_addr.ip4.s_addr = sin6->sin6_addr.s6_addr32 [3];
        from_addr.i4_addr.port = sin6->sin6_port;
      }
      add_from = 1;
    } else if (from->sa_family == AF_INET6) {
      from_addr.discrim = ALLNET_IP6_ADDR_ID;
      from_addr.i6_addr.ip6 = sin6->sin6_addr;
      from_addr.i6_addr.port = sin6->sin6_port;
      from_addr.i6_addr.scope = sin6->sin6_scope_id;
      add_from = 1;
    }
    if (add_from) {
      printf ("   actually, adding sender: ");
      print_addr (&from_addr, 1);
      struct allnet_peer real_peer;
      real_peer.provides = hello->sender.provides;
      real_peer.peer = from_addr;
      add_peer (&real_peer);
      received_now (&real_peer);
    } else {
      add_peer (&(hello->sender));
      received_now (&(hello->sender));
    }
    int i;
    for (i = 0; i < hello->num_peers; i++) {
      if (! (is_my_addr (&(hello->peers [i].peer)))) {  /* not me */
        printf ("adding peer: ");
        print_addr (&(hello->peers [i].peer), 1);
        add_peer (hello->peers + i);
      } else {
        printf ("received my own address from a peer: ");
        print_addr (&(hello->peers [i].peer), 1);
      }
    }
    send_hellos ();  /* added some peer, send hellos if time is right */
  } else {
    printf ("received my own hello message from ");
    print_addr (&(hello->sender.peer), 1);
  }
}

static void recv_message (char * buffer, int size, struct sockaddr * from,
                          int is_local)
{
  struct allnet_header header;
  header.route.addrs = NULL;
  int offset = parse_header (buffer, size, &header);
  if (offset < 0) {
    printf ("received message with invalid header\n");
    return;
  }
  if (header.total_length != size) {
    printf ("incoming message size %d, but from header %lld\n", size,
            header.total_length);
    return;
  }
  int msg_type = header.allnet_type;
  allnet_data_len data_len, sig_len;
  char * data, sig;
  int parse_size;
  allnet_data_len data_size;
  int success, priority;
  struct allnet_hello * hello = NULL;
  switch (msg_type) {
    case ALLNET_TYPE_NONE:
      break;
    case ALLNET_TYPE_HELLO:
#define MAX_RECEIVE_PEERS 1000  /* not much harm in making it big */
      hello = malloc (sizeof (struct allnet_hello) +
                      sizeof (struct allnet_peer) * MAX_RECEIVE_PEERS);
      if (hello == NULL) {
        printf ("unable to allocate hello message\n");
        exit (1);
      }
      parse_size = parse_hello (buffer + offset, size - offset,
                                hello, MAX_RECEIVE_PEERS);
#undef MAX_RECEIVE_PEERS
      printf ("got hello message with %d addresses in route, %d used\n",
              b1 (header.route.num_addrs), b1 (header.route.used));
      print_route (&(header.route), 1);
      print_hello (hello, 1);
      save_hello (hello, from, &header.route);
      free (hello);
      break;
    case ALLNET_TYPE_DATA:   /* forward the data on all interfaces */
    case ALLNET_TYPE_CLEAR:  /* forward the data on all interfaces */
      data_size = header.total_length - offset;
      data = buffer + offset;
      priority = compute_priority (data, data_size, from, is_local);
      forward_packet (&(header.route), data, data_size, priority,
                      (msg_type == ALLNET_TYPE_CLEAR));
      break;
    case ALLNET_TYPE_SIGNED: /* forward the data on all interfaces */
      
    default:
      printf ("received message with unknown type %d\n", msg_type);
      return;
  }
}

/* This implementation is multithreaded, with one thread listening on each
 * socket, and one thread sending to each socket.  There is a per-socket
 * sending queue. Synchronization is used when messages are added to
 * or removed from the queue used for sending on each socket.
 */

/* returns the number of characters read, 0 for EOF, -1 for error */
/* gets rid of initial spaces and anything after the comment char '#' */
static int my_read_line (int fd, char * buffer, int size)
{
  int offset = 0;
  int discarding = 0;
  while (offset + 1 < size) {
    int chars = read (fd, buffer + offset, 1);
    if (chars != 1) {
      buffer [offset] = '\0';
      if (chars == 0) {   /* end of file */
        return 0;
      } else {            /* error */
        perror ("read allnetd.conf");
        return -1;
      }
    }
    if (buffer [offset] == '\n') {
      buffer [offset] = '\0';
      if (offset > 0)
        return offset;
    } else if (discarding) {  /* do nothing */
    } else if (buffer [offset] == '#') {
      discarding = 1;
      buffer [offset] = '\0';
    } else if ((offset == 0) && (buffer [offset] == ' ')) {  /* do nothing */
    } else {
      offset++;      /* normal case */
    }
  }
  buffer [offset] = '\0';
  printf ("line too long\n");
  return -1;
}

static void set_value (char * string, char * key, int * value)
{
  int slen = strlen (string);
  int klen = strlen (key);
  if ((slen > klen) && (strncasecmp (key, string, klen) == 0)) {
    string += slen;
    while ((*string == ':') || (isblank (*string))) {
      string++;
      slen--;
    }
    char * finish;
    int input = strtol (string, &finish, 10);
    if (finish != string) /* conversion success */
      *value = input;
  }
}

static void set_config (char * string,
                        int * Bps_wifi, int * Bps_internet, int * kB)
{
  set_value (string, "Bps-wifi", Bps_wifi);
  set_value (string, "Bps-inet", Bps_internet);
  set_value (string, "kB", kB);
}

/* 2 thread functions.  If all goes well, each should loop forever */
void * listen_loop (void * arg)
{
  int interface = * ((int *) arg);
  free (arg);
  printf ("listen thread started on interface %d\n", interface);
  struct send_queue_info * q = send_queues + interface;
  int sock = q->ip6_udp_socket;
  if (sock < 0)
    sock = q->ip4_udp_socket;
  if (sock < 0) {
    printf ("listen thread unable to find socket on interface %d\n",
            interface);
    print_send_queues ();
    exit (1);
  }
  char * buffer = malloc (MAX_RECEIVE_BUFFER);
  if (buffer == NULL) {
    perror ("malloc in local_listen");
    return NULL;
  }
  struct sockaddr_storage sa;
  struct sockaddr * sap = (struct sockaddr *) (&sa);
  struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) (&sa);
  struct sockaddr_in * sin = (struct sockaddr_in *) (&sa);
  while (1) {
    printf ("sizeof sockaddr = %d, storage = %d, in6 = %d, in = %d\n",
            sizeof (struct sockaddr), sizeof (struct sockaddr_storage),
            sizeof (struct sockaddr_in6), sizeof (struct sockaddr_in));
    socklen_t adr_size = sizeof (sa);
    printf ("calling recvfrom (socket %d, interface %d)\n", sock, interface);
    ssize_t received = recvfrom (sock, buffer, MAX_RECEIVE_BUFFER, 0,
                                 sap, &adr_size);
    if (received < 0)
      perror ("recvfrom");
    printf ("socket %d, interface %d, received %d bytes, address %d bytes\n",
            sock, interface, received, adr_size);
    char from_string [100];
    inet_name (sap, from_string, sizeof (from_string));
    printf ("from %s", from_string);
    if (received >= 0) {
      printf (": ");
      print_buffer (buffer, received, 0);
    }
    printf ("\n");
    if (received > 0)
      recv_message (buffer, received, sap, q->is_local);
    if (q->is_local) {  /* add the port */
      int i;
      int found = 0;
      struct sockaddr_in  * sin  = (struct sockaddr_in  *) sap;
      struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) sap;
      int port_number = sin->sin_port;
      if (sap->sa_family == AF_INET6) 
        port_number = sin6->sin6_port;
      for (i = 0; i < q->num_local_clients; i++) {
        if (q->local_ports [i] == port_number)
          found = 1;
      }
      if (! found) {
        if (q->num_local_clients >= MAX_LOCAL_CLIENTS) {
          printf ("error: too many local clients!\n");
          exit (1);
        }
        printf ("adding local port number %d\n", port_number);
        q->local_ports [q->num_local_clients] = port_number;
        q->num_local_clients++;
      }
    }
  }
  free (buffer);
  return NULL;
}

void * send_loop (void * arg)
{
  int interface = * ((int *) arg);
  free (arg);
  printf ("send thread started on interface %d\n", interface);
  struct send_queue_info * q = send_queues + interface;

  int sock = q->ip6_udp_socket;
  if (sock < 0)
    sock = q->ip4_udp_socket;
  if (sock < 0) {
    printf ("listen thread unable to find socket on interface %d\n",
            interface);
    print_send_queues ();
    exit (1);
  }


  time_t start_time = time (NULL);

  while (1) {
    int buffer_size;
    int buffer_priority;   /* we don't really care, but we might some day */

    /* wait until we have a message */
    printf ("send_loop requesting message from queue %d\n", interface);
    char * buffer = queue_wait (q->queue, &buffer_size, &buffer_priority);
    printf ("send_loop got %d bytes from queue %d\n", buffer_size, interface);

    struct sockaddr_storage sa [SENT_PEERS];
    int found = 0;
    if (q->is_local) {
      found = SENT_PEERS;
      if (found > q->num_local_clients)
        found = q->num_local_clients;
      static int random = 0;

      int i;
      for (i = 0; i < found; i++) {
        struct sockaddr_in  * sin  = (struct sockaddr_in  *) (sa + i);
        sin->sin_family = AF_INET;
        sin->sin_addr.s_addr = htonl (0x7f000001);
        random = (random + 1) % q->num_local_clients;
        sin->sin_port = q->local_ports [random];
      }
    } else {
      found = addresses_from_buffer (buffer, buffer_size, sa, SENT_PEERS,
                                     interface);
    }
    printf ("%d addresses from buffer or local ports\n", found);
    if (found > 0) {   /* we can send the buffer */
      int i;
      for (i = 0; i < found; i++) {
        socklen_t adr_size = sizeof (struct sockaddr_storage);
        int flags = MSG_DONTWAIT | MSG_NOSIGNAL;

 /* wait if needed to make sure we do not exceed the maximum send speed */
        if (q->max_bytes_per_sec > 0) {
          time_t finish_time = start_time + buffer_size / q->max_bytes_per_sec;
          if (time (NULL) < finish_time)
            printf ("waiting to send until time %d, now %d\n",
                    ((int) finish_time), ((int) time (NULL)));
          while (time (NULL) < finish_time)
            sleep (1);
        }

        struct sockaddr * sap = (struct sockaddr *) (sa + i);
        char string [100];
        inet_name (sap, string, sizeof (string));
        printf ("sendto [%d] (%d, %p, %d, %x, %p/%s, %d)\n",
                i, sock, buffer, buffer_size, flags, sap, string, adr_size);
        print_buffer (buffer, buffer_size, 1);
        ssize_t sent = sendto (sock, buffer, buffer_size, flags, sap, adr_size);

        if (sent < 0)
          perror ("sendto");

        time (&start_time);  /* restart the clock */

        printf ("sent %d bytes (result %d) to destination ", buffer_size, sent);
        char to_string [100];
        inet_name (sap, to_string, sizeof (to_string));
        printf ("%s\n", to_string);
      }
    }

    free (buffer);
    start_time = time (NULL);
  }

  return NULL;
}

static int is_loopback (char * name)
{
  return ((name [0] == 'l') && (name [1] == 'o'));
}

static void init_local_interface (int interface, struct ifaddrs * ifa)
{
  if (! is_loopback (ifa->ifa_name))
    return;
  
}

/* returns the queue number if successful, -1 otherwise */
static int init_ipv6_interface (int interface, struct ifaddrs * ifa,
                                int qsize, int Bps)
{
  struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) ifa->ifa_addr;

  if (sin6->sin6_family != AF_INET6)  /* not IPv6, ignore */
    return -1;

/*
  char string [100];
  inet_name (ifa->ifa_addr, string, sizeof (string));
  printf ("first two bytes of ipv6 address %s are %x\n", string,
          sin6->sin6_addr.s6_addr16 [0]);
*/
      /* is it a link-local address or a site-local address? */
  if (((ntohs (sin6->sin6_addr.s6_addr16 [0]) & 0xffc0) == 0xfe80) ||
      ((ntohs (sin6->sin6_addr.s6_addr16 [0]) & 0xffc0) == 0xfec0))
      /* not routable IPv6 addresses, fall back on IPv4 */
    return -1;

  /* a copy of the address passed in, so we can set the port number */
  struct sockaddr_storage sa;
  struct sockaddr_in6 * sap = (struct sockaddr_in6 *) (&sa);
  *sap = *sin6;
  sap->sin6_port = ALLNET_PORT;

  int q;
  int sock = socket (AF_INET6, SOCK_DGRAM, 0);
  struct sockaddr_in6 baddr;
  bzero (&baddr, sizeof (baddr));
  baddr.sin6_family = AF_INET6;
  baddr.sin6_port = ALLNET_PORT;
  if (is_loopback (ifa->ifa_name)) {
    baddr.sin6_addr.s6_addr32 [3] = 1;  /* loopback address */
    baddr.sin6_port = ALLNET_LOCAL_PORT;
  }
  if (bind (sock, ((struct sockaddr *) (&baddr)), sizeof (baddr)) != 0) {
    printf ("ipv6 UDP port %x on interface %s already taken\n",
            ntohs (baddr.sin6_port), ifa->ifa_name);
    return -1;
  }
  if (is_loopback (ifa->ifa_name))  /* no max size, no speed limit */
    q = new_send_queue (ifa->ifa_name, -1, 0, 1, 0, -1, -1, sock, -1, -1,
                        0, NULL, 0, NULL);
  else
    q = new_send_queue (ifa->ifa_name, -1, qsize, 0, Bps, -1, -1, sock, -1, -1,
                        0, NULL, 1, (struct sockaddr *) sap);
  printf ("initialized %sipv6 interface %s %d, if number %d\n",
          (is_loopback (ifa->ifa_name)) ? "loopback " : "", ifa->ifa_name, q,
          interface);
  return q;
}

/* returns the queue number if successful, -1 otherwise */
static int init_ipv4_interface (int interface, struct ifaddrs * ifa,
                                int qsize, int Bps)
{
  struct sockaddr_in  * sin  = (struct sockaddr_in  *) ifa->ifa_addr;
  if (sin->sin_family != AF_INET)
    return -1;

  /* a copy of the address passed in, so we can set the port number */
  struct sockaddr_storage sa;
  struct sockaddr_in * sap = (struct sockaddr_in *) (&sa);
  *sap = *sin;
  sap->sin_port = ALLNET_PORT;

  int q;
  int sock = socket (AF_INET, SOCK_DGRAM, 0);
  struct sockaddr_in baddr;
  bzero (&baddr, sizeof (baddr));
  baddr.sin_family = AF_INET;
  baddr.sin_port = ALLNET_PORT;
  if (is_loopback (ifa->ifa_name))
    baddr.sin_port = ALLNET_LOCAL_PORT;
  if (bind (sock, ((struct sockaddr *) (&baddr)), sizeof (baddr)) != 0) {
    printf ("ipv4 UDP port %x on interface %s already taken\n",
            ntohs (baddr.sin_port), ifa->ifa_name);
    return -1;
  }
  
  char string [100];
  inet_name ((struct sockaddr *) sap, string, sizeof (string));
  printf ("ipv4 init for interface %d, address %s\n", interface, string);
  if (is_loopback (ifa->ifa_name))  /* no max size, no speed limit */
    q = new_send_queue (ifa->ifa_name, -1, 0, 1, 0, -1, sock, -1, -1, -1,
                        0, NULL, 0, NULL);
  else
    q = new_send_queue (ifa->ifa_name, -1, qsize, 0, Bps, -1, sock, -1, -1, -1,
                        0, NULL, 1, (struct sockaddr *) sap);
  printf ("initialized %sipv4 interface %s %d, if number %d\n",
          (is_loopback (ifa->ifa_name)) ? "loopback " : "", ifa->ifa_name, q,
          interface);
  return q;
}

/* returns the queue number if successful, -1 otherwise */
static int init_allnet_interface (int interface, struct ifaddrs * ifa,
                                  int qsize, int Bps)
{
  struct sockaddr_ll * sll  = (struct sockaddr_ll  *) ifa->ifa_addr;
  if (sll->sll_family != AF_PACKET)
    return -1;

/* need to execute the commands:
      sudo iw dev $if set type ibss
      sudo ifconfig $if up
      sudo iw dev $if ibss join allnet 2412
 */
  static char command [1000];
  snprintf (command, sizeof (command), "iw dev %s set type ibss",
            ifa->ifa_name);
  int sys_result = system (command);
  printf ("result of calling '%s' was %d\n", command, sys_result);
  if (sys_result != 0) {
    if (sys_result != -1)
      printf ("program exit status for %s was %d\n",
              command, WEXITSTATUS (sys_result));
    printf ("probably a wired or configured interface\n");
    return -1;
  }

  snprintf (command, sizeof (command), "ifconfig %s up", ifa->ifa_name);
  sys_result = system (command);
  printf ("result of calling '%s' was %d\n", command, sys_result);
  if (sys_result != 0) {
    if (sys_result != -1)
      printf ("program exit status for %s was %d\n",
              command, WEXITSTATUS (sys_result));
    return -1;
  }

  snprintf (command, sizeof (command), "iw dev %s ibss join allnet 2412",
            ifa->ifa_name);
  sys_result = system (command);
  printf ("result of calling '%s' was %d\n", command, sys_result);
  if (sys_result != 0) {
    if ((sys_result != -1) && (WEXITSTATUS (sys_result) == 142)) {
      printf ("interface is already connected to allnet\n");
    } else if (sys_result != -1) {
      printf ("program exit status for %s was %d\n",
              command, WEXITSTATUS (sys_result));
      return -1;
    }
  } else {
    printf ("success!!! configured wireless interface %s\n",
            ifa->ifa_name);
  }

/* should accomodate old-style interfaces that only support iwconfig */
  int packet_socket = socket (AF_PACKET, SOCK_DGRAM,
                              ALLNET_WIFI_PROTOCOL);
  if (packet_socket < 0) {
    perror ("packet socket");
    return -1;
  }
  struct sockaddr_ll bind_addr;
  bind_addr.sll_family = AF_PACKET;
  bind_addr.sll_protocol = ALLNET_WIFI_PROTOCOL;
  bind_addr.sll_ifindex = sll->sll_ifindex;
  if (bind (packet_socket, (struct sockaddr *) (&bind_addr),
    sizeof (bind_addr)) < 0) {
    perror ("bind for packet socket");
    return -1;
  }
  struct allnet_addr my_addr;
  my_addr.discrim = ALLNET_UNICAST_ADDR_ID;
  my_addr.uc_addr.hops [0] = 1;  /* not meaningful */
  my_addr.uc_addr.num_bytes [0] = sll->sll_halen;
  if (sll->sll_halen > sizeof (my_addr.uc_addr.id))
    my_addr.uc_addr.num_bytes [0] = sizeof (my_addr.uc_addr.id);
  memcpy (my_addr.uc_addr.id, sll->sll_addr, b1 (my_addr.uc_addr.num_bytes));
  int q = new_send_queue (ifa->ifa_name, sll->sll_ifindex, qsize, 0, Bps,
                          packet_socket, -1, -1, -1, -1, 1, &my_addr, 0, NULL);
  printf ("initialized allnet interface %s %d, if number %d\n",
          ifa->ifa_name, q, interface);
  return q;
}

#define MAX_INTERFACES	100
/* returns -1 if not found, or the interface position if found */
static int interface_number (char * name,
                             char interfaces [MAX_INTERFACES] [IFNAMSIZ])
{
  int i;
  for (i = 0; i < MAX_INTERFACES; i++)
    if (strcmp (name, interfaces [i]) == 0)
      return i;
  return -1;
}

static void init_interfaces (int max_Bps_allnet, int max_Bps_internet,
                             int total_bytes)
{
  init_send_queues ();
#define MAX_INTERFACES	100
  char interfaces [MAX_INTERFACES] [IFNAMSIZ];
  int queue_number [MAX_INTERFACES];
  int i;
  int found;
  for (i = 0; i < MAX_INTERFACES; i++)
    interfaces [i] [0] = '\0';   /* empty string, "" */
  struct ifaddrs * ifa;
  if (getifaddrs (&ifa) != 0) {
    perror ("getifaddrs");
    exit (1);
  }
  struct ifaddrs * ifa_loop = ifa;
  int interfaces_found = 0;
  int nonlocal_interfaces_found = 0;
  while (ifa_loop != NULL) {
    struct sockaddr_ll  * sll  = (struct sockaddr_ll  *) ifa_loop->ifa_addr;
    struct sockaddr_in  * sin  = (struct sockaddr_in  *) ifa_loop->ifa_addr;
    struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) ifa_loop->ifa_addr;
    if (sll->sll_family == AF_PACKET) {
      interfaces_found++;
      if (! (is_loopback (ifa_loop->ifa_name))) 
        nonlocal_interfaces_found++;
    }
    if (sin->sin_family == AF_INET) {
      struct sockaddr_in  * copy =
         (struct sockaddr_in  *) (my_inet_addrs + num_my_inet_addrs);
      *copy = *sin;
      num_my_inet_addrs++;
    }
    if (sin->sin_family == AF_INET6) {
      struct sockaddr_in6 * copy =
         (struct sockaddr_in6 *) (my_inet_addrs + num_my_inet_addrs);
      *copy = *sin6;
      num_my_inet_addrs++;
    }
    ifa_loop = ifa_loop->ifa_next;
  }
  for (i = 0; i < num_my_inet_addrs; i++) {
    char string [100];
    inet_name ((struct sockaddr *) (my_inet_addrs + i),
                string, sizeof (string));
    printf ("my_inet_addrs [%d] is %s\n", i, string);
  }
  /* we distribute the space evenly among the nonlocal interfaces */
  int bytes_each = total_bytes;
  if (nonlocal_interfaces_found > 0)
    bytes_each = total_bytes / nonlocal_interfaces_found;
  ifa_loop = ifa;
  int count = 0;
  while (ifa_loop != NULL) {
    printf ("interface %s, %d of %d (%d nonlocal), %d bytes, ",
            ifa_loop->ifa_name, ++count, interfaces_found,
            nonlocal_interfaces_found, bytes_each);
    char address_string [100];
    inet_name (ifa_loop->ifa_addr, address_string, sizeof (address_string));
    printf ("address %s\n", address_string, sizeof (address_string));
    struct sockaddr_ll  * sll  = (struct sockaddr_ll  *) ifa_loop->ifa_addr;
    if (sll->sll_family == AF_PACKET) {
      int interface = interface_number (ifa_loop->ifa_name, interfaces);
      if ((interface == -1) && (sll->sll_ifindex < MAX_INTERFACES)) {
        strncpy (interfaces [sll->sll_ifindex], ifa_loop->ifa_name, IFNAMSIZ);
        queue_number [sll->sll_ifindex] = -1;
      }
    }
    int interface = interface_number (ifa_loop->ifa_name, interfaces);
    /* first try to open it as an IPv6 interface, which includes IPv4 */
    if (interface != -1)
      queue_number [interface] =
        init_ipv6_interface (interface, ifa_loop,
                             bytes_each, max_Bps_internet);
    ifa_loop = ifa_loop->ifa_next;
  }

  /* now try to open any remaining interfaces as IPv4 interfaces */
  ifa_loop = ifa;
  while (ifa_loop != NULL) {
    int interface = interface_number (ifa_loop->ifa_name, interfaces);
    if ((interface != -1) && (queue_number [interface] == -1))
      queue_number [interface] =
        init_ipv4_interface (interface, ifa_loop, 
                             bytes_each, max_Bps_internet);
    ifa_loop = ifa_loop->ifa_next;
  }
  for (i = 0; i < MAX_INTERFACES; i++)
    if ((strcmp (interfaces [i], "") != 0) && (queue_number [i] == -1))
      printf ("interface %d (%s) still unconfigured\n", i, interfaces [i]);

  /* finally, try to open any unconfigured interfaces as Allnet wireless */
  ifa_loop = ifa;
  while (ifa_loop != NULL) {
    int interface = interface_number (ifa_loop->ifa_name, interfaces);
    if ((interface != -1) && (queue_number [interface] == -1))
      queue_number [interface] =
          init_allnet_interface (interface, ifa_loop,
                                 bytes_each, max_Bps_allnet);
    ifa_loop = ifa_loop->ifa_next;
  }
  freeifaddrs (ifa);
  print_send_queues ();
}

main (int argc, char ** argv)
{
  int max_Bps_wifi = 1250;      /* 10 Kbit/s = 1250 bytes per second*/
  int max_Bps_internet = 12500; /* 100 Kbit/s = 1250 bytes per second*/
  int max_kB = 10000;           /* 10MB of memory usage */

  /* check config file */
  int cfd = open ("allnetd.conf", O_RDONLY);
  if (cfd != -1) {
    static char conf_buf [1000];
    while (my_read_line (cfd, conf_buf, sizeof (conf_buf)) > 0) {
      set_config (conf_buf, &max_Bps_wifi, &max_Bps_internet, &max_kB);
    }
    close (cfd);
  }
  printf ("allnetd: max Bps %d on wifi, ", max_Bps_wifi);
  printf ("%d on internet, ", max_Bps_internet);
  printf ("at most %d KB of memory\n", max_kB);

  /* open all the interfaces, including one listening for local traffic */
  /* initialize the send_queues array */
  /* create a send thread for each send_queue */
  /* create a listen thread for each interface */
  init_interfaces (max_Bps_wifi, max_Bps_internet, max_kB * 1000);

  /* for testing, add 128.171.10.118 as a peer */
  if (argc == 1) {
    struct allnet_addr test_addr;
    test_addr.discrim = ALLNET_IP4_ADDR_ID;
    inet_pton (AF_INET, "128.171.10.118", ((void *) &(test_addr.i4_addr)));
    test_addr.i4_addr.port = ALLNET_PORT;
    print_addr (&test_addr, 1);
    struct allnet_peer test_peer;
    test_peer.peer = test_addr;
    test_peer.provides = 0;
    add_peer (&test_peer);

    send_hellos ();

    sleep (2);
    int size = 10000;
    char * buffer = malloc (size);
    struct allnet_addr bcast;
    bcast.discrim = ALLNET_BCAST_ADDR_ID;
    bcast.bc_addr.hops [0] = 5;
    struct allnet_route route;
    route.num_addrs [0] = 1;
    route.used [0] = 0;
    route.addrs = &bcast;
    int sending = store_data ("hello world", strlen ("hello world") + 1, 1,
                              &route, buffer, size);
    printf ("sending %d bytes of hello world\n", sending);
    queue_add (send_queues [1].queue, buffer, sending, HELLO_PRIORITY);
  }

  /* loop forever */
  time_t start = time (NULL);
  while (1) {
    sleep (HELLO_INTERVAL);
    send_hellos ();  /* every 30 min */
  }
}

