/* interface to allnet functions */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "types.h"
#include "messages.h"
#include "api.h"

/* allnet_start should be called before using allnet, and allnet_stop
 * should be called after finishing with allnet.
 *
 * allnet_start opens and returns a given port for communication with
 * the allnet daemon.
 *
 * whenever a message is available, the callback function is called.
 * the callback function must complete execution and return before
 * the next message will be delivered.
 *
 * all the parameters passed to the callback function may be reused
 * by allnet on the next call, so should not be saved anywhere.
 *
 * the callback function may call allnet_stop or allnet_send
 *
 * the caller may provide an arbitrary pointer ("arg") which will
 * be given to the callback function on each call.
 *
 * the return value from allnet_start is a port number >= 0 (success),
 * otherwise -1 (failure).
 */

struct port_info {
  int in_use;
  int socket;
  allnet_receive_fun callback;
  void * callback_arg;
  char * buffer;
  int bsize;
  struct allnet_addr addr;
  struct allnet_route route;
  pthread_t receive_thread;
  int stop_receiving;
};

#define MAX_PORTS	10
struct port_info open_ports [MAX_PORTS];
int initialized = 0;

void * receive_loop (void * arg);

/*
typedef void (* allnet_receive_fun) (void * arg, char * buffer, int bsize,
                                     struct allnet_addr * from,
                                     struct allnet_route * route, int key_id);
*/
int allnet_start (allnet_receive_fun callback, void * arg)
{
  if (! initialized) {
    int i;
    for (i = 0; i < MAX_PORTS; i++)
      open_ports [i].in_use = 0;
    initialized = 1;
  }

  int index = 0;
  while ((index < MAX_PORTS) && (open_ports [index].in_use))
    index++;
  if (index >= MAX_PORTS) {
    printf ("no more allnet ports available\n");
    return -1;
  }

  struct port_info * p = open_ports + index;

/* do the easy initializations first */
  p->callback = callback;
  p->callback_arg = arg;

  p->addr.discrim = ALLNET_NO_ADDR_ID;

  p->route.num_addrs [0] = 0;
  p->route.addrs = NULL;

  p->stop_receiving = 0;

  p->socket = socket (AF_INET, SOCK_DGRAM, 0);
  if (p->socket < 0) {
    perror ("socket");
    return -1;
  }

  p->bsize = 66000;
  p->buffer = malloc (p->bsize);
  if (p->buffer == NULL) {
    printf ("unable to create receive buffer\n");
    return -1;
  }

  if (pthread_create (&(p->receive_thread), NULL,
                      receive_loop, (void *) index) != 0) {
    free (p->buffer);
    printf ("unable to create receive thread\n");
    return -1;
  }

  /* send an empty message to the allnet daemon to let them know we are here */
  struct sockaddr_in sin;
  sin.sin_family = AF_INET;
  /* set the loopback address */
  sin.sin_addr.s_addr = htonl (0x7F000001);
  sin.sin_port = ALLNET_LOCAL_PORT;
  /* should we even check the return value? */
  if (sendto (p->socket, p->buffer, 0, 0,
              (struct sockaddr *) (&sin), sizeof (sin)) < 0)
    perror ("initial sendto");

  /* commit the new port */
  p->in_use = 1;

  return index + 1;
}

void allnet_stop (int port)
{
  if ((port > 0) && (port <= MAX_PORTS))
    open_ports [port - 1].stop_receiving = 1;
/* the receive thread does all the shutting down before exiting */
}

void check_end_thread (int port)
{
  struct port_info * p = open_ports + (port - 1);
  if (p->stop_receiving) {
    close (p->socket);
    free (p->route.addrs);
    free (p->buffer);
    p->in_use = 0;
    pthread_exit (NULL);
  }
}

void * receive_loop (void * arg) {
  int index = (int) arg;
  struct port_info * p = open_ports + index;
  int flags = MSG_DONTWAIT | MSG_TRUNC;
  struct sockaddr_storage sa;
  struct sockaddr * sap = (struct sockaddr *) &sa;
  while (1) {
    socklen_t al = sizeof (sa);
#ifdef DEBUG_PACKETS
    printf ("receive_loop on port %d (%d, %p, %d, %x, %p, %d)\n",
            index, p->socket, p->buffer, p->bsize, flags, sap, al);
#endif /* DEBUG_PACKETS */
    int received;
    do {
      received = recvfrom (p->socket, p->buffer, p->bsize, flags, sap, &al);
      if ((received == -1) && ((errno == EAGAIN) || (errno == EWOULDBLOCK))) {
        received = -2;
        struct timespec t;
        t.tv_sec = 0;
        t.tv_nsec = 10 * 1000 * 1000;   /* sleep 10ms or so */
        nanosleep (&t, NULL);
      }
    } while (received == -2);
    if (received == -1) {
      perror ("recvfrom");
      continue;   /* receive again */
    }
#ifdef DEBUG_PACKETS
    printf ("receive_loop on port %d received %d bytes\n", index, received);
#endif /* DEBUG_PACKETS */
    check_end_thread (index);
    if (received > p->bsize) {
      printf ("received message had %d bytes, last %d bytes lost\n", received,
              received - p->bsize);
      received = p->bsize;
    }
    struct allnet_header header;
    int offset = parse_header (p->buffer, received, &header);
    if (offset < 0) {
      printf ("unable to parse received buffer\n");
      continue;   /* receive again */
    }
    p->addr = header.route.addrs [b1 (header.route.used)];
    p->route = header.route;
    int size = received - offset;
    char * data = p->buffer + offset;
    int call_app = 1;

    /* try to decrypt the packet */
    char * key_hint = NULL;
    int chosen_key_id = 0;
    if (p->addr.discrim == ALLNET_KEY_ADDR_ID) {
      int hint_size = b1 (p->addr.pk_addr.num_bits) / 8;
      if (hint_size > sizeof (p->addr.pk_addr.pubkey))
        hint_size = sizeof (p->addr.pk_addr.pubkey);
      if (hint_size > 0)
        key_hint = p->addr.pk_addr.pubkey;
      char * text_buffer = malloc (size);
      if (text_buffer != NULL) {
        int decoded = 
          decrypt_verify (data, size, text_buffer, size, key_hint, hint_size,
                          &chosen_key_id);
        if (decoded > size) {
          printf ("decoded message %d larger than received %d\n",
                  decoded, size);
          decoded = size;
        }
        if (decoded == -1) {
          /* too many to print, including my own messages sent back to me
           printf ("unable to decrypt incoming message, discarding\n"); */
          call_app = 0;
        } else {
          memcpy (data, text_buffer, decoded);
          size = decoded;
        }
        free (text_buffer);
      }
    }

    if (call_app)
      p->callback (p->callback_arg, data, size,
                   header.route.addrs + b1 (header.route.used),
                   &(header.route), chosen_key_id);
#ifdef DEBUG_PACKETS
    if (call_app)
      printf ("callback done, argument %d\n", index);
#endif /* DEBUG_PACKETS */
    check_end_thread (index);
  }
}

/* sends the message.
 *
 *  If the address is NULL, the message will be sent using
 * a broadcast, otherwise using the given address.
 *
 * If the key is 0, the message will be sent in the clear,
 * otherwise sent encrypted with the given key.
 */
void allnet_send (int port, char * message, int mlen,
                  struct allnet_addr * dest, int key_id)
{
  if ((port <= 0) || (port > MAX_PORTS))
    return;
  struct port_info * p = open_ports + (port - 1);

  struct allnet_addr to;
  if (dest != NULL) {
    to = *dest;
  } else {
    if (key_id == 0) {
      to.discrim = ALLNET_BCAST_ADDR_ID;
      to.bc_addr.hops [0] = 6;
    } else {
      to.discrim = ALLNET_KEY_ADDR_ID;
      to.pk_addr.hops [0] = 6;
      to.pk_addr.num_bits [0] = 0;
    }
  }

  struct allnet_route route;
  route.num_addrs [0] = 1;
  route.used [0] = 0;
  route.addrs = &to;

  char cypher [2048];
  int clen = sizeof (cypher);
  if (key_id != 0) {
    int s = sign_encrypt (message, mlen, cypher, clen, key_id, NULL, 0);
    if (s == -1) {
      printf ("unable to encrypt message '%s'\n", message);
      return;
    }
    mlen = s;
    message = cypher;
  }

  char * buffer = "";
  /* get the size first */
  int size = - store_data (message, mlen, (key_id == 0), &route, buffer, 0);
  buffer = malloc (size);
  if (buffer == NULL) {
    printf ("unable to allocate buffer for message to send, sizes %d/%d\n",
            clen, size);
    return;
  }
  /*
  printf ("using buffer of size %d to send %d bytes of user data\n",
          size, mlen);
  */

  if (store_data (message, mlen, (key_id == 0), &route, buffer, size) != size) {
    printf ("error: unknown result from store_data\n");
    exit (1);
  }

  struct sockaddr_in sin;
  sin.sin_family = AF_INET;
  /* set the loopback address */
  sin.sin_addr.s_addr = htonl (0x7F000001);
  sin.sin_port = ALLNET_LOCAL_PORT;
  if (sendto (p->socket, buffer, size, 0,
              (struct sockaddr *) (&sin), sizeof (sin)) < 0)
    perror ("sendto");
}

