/* interface to allnet functions */

#ifndef ALLNET_API_H
#define ALLNET_API_H

#include "types.h"

/* allnet_start should be called before using allnet, and allnet_stop
 * should be called after finishing with allnet.
 *
 * allnet_start opens and returns a given port for communication with
 * the allnet daemon.
 *
 * whenever a message is available, the callback function is called.
 * the callback function must complete execution and return before
 * the next message will be delivered.
 *
 * all the parameters passed to the callback function may be reused
 * by allnet on the next call, so should not be saved anywhere.
 *
 * the callback function may call allnet_stop or allnet_send
 *
 * the caller may provide an arbitrary pointer ("arg") which will
 * be given to the callback function on each call.
 *
 * the return value from allnet_start is a port number >= 0 (success),
 * otherwise -1 (failure).
 */
typedef void (* allnet_receive_fun) (void * arg, char * buffer, int bsize,
                                     struct allnet_addr * from,
                                     struct allnet_route * route, int key_id);
extern int allnet_start (allnet_receive_fun callback, void * arg);
extern void allnet_stop (int port);

/* functions implemented in allnet_api.c */
/* sends the message.
 *
 *  If the address is NULL, the message will be sent using
 * a broadcast, otherwise using the given address.
 *
 * If the key is 0, the message will be sent in the clear,
 * otherwise sent encrypted with the given key.
 */
extern void allnet_send (int port, char * message, int mlen,
                         struct allnet_addr * dest, int key_id);

/* functions for key management, implemented in allnet_key.c */

/* returns the number n of public keys if successful, otherwise -1.
 *
 * if the file did not exist or was empty, returns 0, but can use
 * keys if they are added with add_key.
 *
 * key IDs normally have values from 1..n */
extern int init_key_file (char * fname, char * myname);

/* returns 1 if it found a name for the given key.  Puts as much of the
 * name as possible into namebuf (and always terminates the string, as
 * long as bsize > 0).  If not found, returns 0, -1 for errors.
 * if a key does not have a name, namebuf is set to "",
 * but name_for_key still returns 1.
 */
extern int name_for_key (int key_id, char * namebuf, int bsize);

/* given any part of a name, returns all the matching key IDs, up to
 * num_keys.  Also returns the total number of matching keys, which
 * may be greater than num_keys.  Returns -1 in case of errors.
 */
extern int key_for_name (char * name, int * key_id, int num_keys);

/* adds the given key, either temporarily (persist = 0) or permanently
 * (persist != 0, stored in the file).
 * returns the key_id of the new key, or -1 in case of failure.
 */
extern int add_key (char * name, char * key, int key_len, int persist);

/* fills in cypher with the signed and encrypted text.  key_id is the
 * ID identifying the recipient.  Also fills in the first ksize
 * bytes of the key, padding with zeros if ksize > key length.
 * returns the number of bytes in the cyphertext if successful -- if
 * the value returned > csize, the result does not fit and is not
 * stored in cypher.
 * returns -1 in case of failure.
 */
extern int sign_encrypt (char * text, int tsize, char * cypher, int csize,
                         int key_id, char * key, int ksize);

/* if successfully decrypts and verifies, fills in text (up to tsize) and
 * returns the total number of characters in text (which may exceed tsize).
 * key_hint, if not NULL, is used to restrict the number of personal keys
 * which need to be tried.
 *
 * if it finds a matching key, sets *key_id (if not NULL)
 *
 * if unable to find a key to decrypt and verify, returns -1
 */
extern int decrypt_verify (char * cypher, int csize, char * text, int tsize,
                           char * key_hint, int ksize, int * key_id);

/* verifies that the signature matches the text for the given public key.
 * the entire key must be provided.  This is intended for key exchange,
 * where in one phase both sides exchange the signatures of a string known
 * to both, then in the second phase, they send the (hopefully) matching
 * public key.
 *
 * returns 1 if successful, 0 if not matched, -1 for errors.
 */
extern int verify_signature (char * sig, int ssize, char * text, int tsize,
                             char * key, int ksize);

#endif /* ALLNET_API_H */
