
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <openssl/rsa.h>
#include <openssl/pem.h>
#include <openssl/objects.h>
#include <openssl/err.h>

#include "api.h"

/* apparently not always in string.h */
char *strcasestr(const char *haystack, const char *needle);

static int open_create_file (char * fname, off_t * size, FILE * * file)
{
  int mode = S_IRWXU;
  int fd = open (fname, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
  if (fd < 0) {
    perror ("open");
    return -1;
  }
  if (size != NULL) {
    struct stat info;
    if (fstat (fd, &info) < 0) {
      perror ("fstat");
      return -1;
    }
    *size = info.st_size;
  }
  if (file != NULL) {
    *file = fdopen (fd, "a+");
    if (*file == NULL) {
      perror ("fdopen");
      return -1;
    }
  }
  return fd;
}

/* append to the end of a file and print any errors.
 * Returns 1 in case of success, 0 error */
static int my_append (int fd, char * buffer, int bytes, char * description)
{
  if (lseek (fd, 0, SEEK_END) == ((off_t) -1)) {
    perror ("lseek in my_append");
    if (description != NULL)
      printf ("%s seek error", description);
    return 0;
  }
  int result = write (fd, buffer, bytes);
  if (result != bytes) {
    if (result < 0)
      perror ("write");
    if (description != NULL)
      printf ("%s error: ", description);
    printf ("wrote %d bytes, actually wrote %d bytes\n", bytes, result);
    return 0;
  }
  return 1;
}

/* return -1 in case of error, or the number of bytes read otherwise.
 * namep (if not NULL) is set to point to a newly allocated buffer
 * containing the name
 */
static int my_readline (FILE * file, char ** namep)
{
  long start = ftell (file);
  int count = 0;
  while (1) {
    int c = fgetc (file);
    if (c == EOF) { /* no newline found */
      if (namep != NULL)
        *namep = NULL;
      return -1;
    }
    if (c == '\n')
      break;
    count++;
  }
  if ((namep == NULL) || (count == 0))
    return count;
  char * name = malloc (count + 1);
  *namep = name;
  if (name == NULL) {
    printf ("unable to allocate name for key\n");
    return -1;
  }
  fseek (file, start, SEEK_SET);  /** go back and record the characters */
  int items = fread (name, 1, count, file);
  if (items != count) {
    printf ("tried to read %d chars for key name, but read %d\n", count, items);
    return -1;
  }
  name [count] = '\0';
  return count;
}

/* keys are stored as a name, followed by a newline, followed by:
 * - for public keys, the PEM-encoded public key
 * - for private keys, the PEM-encoded private key then the public key
 *
 * public keys matching our own private keys are normally stored in the
 * public key file as well.
 */

/* gets the private key and public key and name.
 * return -1 in case of errors, and otherwise, the number of bytes read.
 */
static int get_keys (FILE * file,
                     RSA ** priv_key, RSA ** pub_key, char ** name)
{
  long start = ftell (file);

  if (my_readline (file, name) <= 0) {
    /* eof, do not print: printf ("error reading key name from file\n"); */
    return -1;
  }
  if (priv_key != NULL) {
    /* if all goes well, creates new RSA */
    *priv_key = PEM_read_RSAPrivateKey (file, NULL, NULL, NULL);
    if (*priv_key == NULL) {
      printf ("error reading private key from file\n");
      return -1;
    }
  }
  *pub_key = PEM_read_RSAPublicKey (file, NULL, NULL, NULL);
  if (*pub_key == NULL) {
    printf ("error reading public key from file\n");
    if (priv_key != NULL)
      RSA_free (*priv_key);
    return -1;
  }

  long finish = ftell (file);
  return (finish - start);
}

/* saves the name, the private (if supplied) key, and the public key.
 * return -1 in case of errors, and otherwise, the number of bytes written.
 */
static int save_key (FILE * file, RSA * priv_key, RSA * pub_key, char * name)
{
  long start = ftell (file);
  fprintf (file, "%s\n", name);
  if (priv_key != NULL) {
    if (! PEM_write_RSAPrivateKey (file, priv_key,
                                   NULL, NULL, 0, NULL, NULL)) {
      printf ("unable to write private key\n");
      return -1;
    }
    printf ("saved private key\n");
  }
  if (! PEM_write_RSAPublicKey (file, pub_key)) {
    printf ("unable to write public key\n");
    return -1;
  }

  long finish = ftell (file);
  return (finish - start);
}

/* returns 1 if the two numbers are the same, 0 otherwise */
static int same_bn (BIGNUM * n1, BIGNUM * n2)
{
  if ((n1 == NULL) && (n2 == NULL))
    return 1;
  if ((n1 == NULL) || (n2 == NULL))
    return 0;
  int result = BN_cmp (n1, n2);
  if (result == 0)
    return 1;
  return 0;
}

/* returns 1 if the public part of the two keys matches, 0 otherwise */
static int same_key (RSA * k1, RSA * k2)
{
  return (same_bn (k1->n, k2->n) && same_bn (k1->e, k2->e));
}

void callback (int type, int count, void * arg)
{
  if (type == 0)
    printf (".");
  if (type == 1)
    printf (",");
  if (type == 2)
    printf ("!");
}

/* called if we have no private keys */
static RSA * generate_private_key ()
{
  int bits = 4096;
  printf ("generating %d-bit private key", bits);
  RSA * result = RSA_generate_key (bits, 65537, callback, NULL);
  printf ("\n");
  return result;
}

static int private_fd = -1;
static int public_fd = -1;
static FILE * private_file = NULL;
static FILE * public_file = NULL;

struct key_info {
  int name_size;
  char * name;
  RSA * pubkey;
  int has_privkey;   /* may be zero, and then privkey would be null */
  RSA * privkey;
};

#define MAX_KEYS	1000
static struct key_info keys [MAX_KEYS];
static int saved_keys = 0;
static int num_private_keys = 0;

/* returns the number n of public keys if successful, otherwise -1.
 * key IDs normally have values from 1..n */
int init_key_file (char * fname, char * myname)
{
#define PRIV_STRING	".private"
  int priv_size = strlen (fname) + strlen (PRIV_STRING) + 1;
  char * priv_name = malloc (priv_size);
  if (priv_name == NULL) {
    printf ("unable to allocate name for private key file\n");
    exit (1);
  }
  snprintf (priv_name, priv_size, "%s%s", fname, PRIV_STRING);
#undef PRIV_STRING
  off_t private_size, public_size;
  private_fd = open_create_file (priv_name, &private_size, &private_file);
  if (private_fd < 0)
    return -1;
  public_fd = open_create_file (fname, &public_size, &public_file);
  if (public_fd < 0)
    return -1;

  int private = 1;   /* read private keys first */
  while (1) {
    if (saved_keys >= MAX_KEYS)
      break;
    RSA * pub_key = NULL;
    RSA * priv_key = NULL;
    char * name = NULL;
    int b;
    if (private)
      b = get_keys (private_file, &priv_key, &pub_key, &name);
    else
      b = get_keys (public_file, NULL, &pub_key, &name);
    if (b <= 0) {
      if (! private)
        break;   /* normal end of the loop -- finished reading public keys */
      if (num_private_keys == 0) {
        printf ("no key found, generating private key for name '%s'\n", myname);
        RSA * key = generate_private_key ();
        save_key (private_file, key, key, myname);
        long location = ftell (public_file);   /* should be zero */
        fseek (public_file, 0, SEEK_END);  /* write the public name too */
        save_key (public_file, NULL, key, myname);
        fseek (public_file, location, SEEK_SET);  /* go back */
        fseek (private_file, 0, SEEK_SET);  /* try to read again */
      } else {
        private = 0;  /* start reading public keys */
      }
    } else { /* b > 0 */
      if (private)
        private_size -= b;
      else
        public_size -= b;
      if (private)
        num_private_keys++;
      int name_size = strlen (name);
      printf ("got ");
      if (private)
        printf ("private (%d) and ", RSA_size (priv_key));
      printf ("public (%d bytes) key for %s/%d\n", RSA_size (pub_key),
              name, name_size);
      int already_have = 0;
      if (! private) {  /* public key, we may already have private key */
        int i;
        for (i = 0; i < num_private_keys; i++) {
          if (same_key (pub_key, keys [i].pubkey)) {
            already_have = 1;    /* matching public key */
            printf ("found public key %s matching private key %s\n", name,
                    keys [i].name);
          }
        }
      }
      if (already_have) {        /* do not save this key */
        if (pub_key != NULL)
          RSA_free (pub_key);
        if (priv_key != NULL)
          RSA_free (priv_key);
        continue;                /* start loop again */
      }
      struct key_info * s = keys + saved_keys;
      s->name_size = name_size;
      s->name = name;
      s->pubkey = pub_key;
      s->has_privkey = private;
      s->privkey = NULL;
      if (private)
        s->privkey = priv_key;
      saved_keys++;
    }
  }
  printf ("got %d keys, of which %d are my own\n",
          saved_keys, num_private_keys);
  return saved_keys - num_private_keys;
}

/* returns 1 if it found a name for the given key.  Puts as much of the
 * name as possible into namebuf.  If not found, returns 0, -1 for errors.
 * if a key does not have a name, namebuf is set to "",
 * but name_for_key still returns 1.
 */
int name_for_key (int key_id, char * namebuf, int bsize)
{
  if ((key_id >= 0) && (key_id < saved_keys)) {
    strncpy (namebuf, keys [key_id].name, bsize - 1);
    namebuf [bsize - 1] = '\0';
    return 1;
  }
  return 0;
}

/* given any part of a name, returns all the matching key IDs, up to
 * num_keys.  Also returns the total number of matching keys, which
 * may be greater than num_keys.  Returns -1 in case of errors.
 */
int key_for_name (char * name, int * key_id, int num_keys)
{
  int result = 0;
  int i;
  for (i = 0; i < saved_keys; i++) {
    if (strcasestr (keys [i].name, name) != NULL) {
      if (num_keys > result)
        key_id [result] = i;
      result++;
    }
  }
  return result;
}

/* adds the given key, either temporarily (persist = 0) or permanently
 * (persist != 0, stored in the file).
 * returns the key_id of the new key, or -1 in case of failure.
 */
int add_key (char * name, char * key, int key_len, int persist)
{
  return -1;
}

/* on each call, initialize the random number generator, needed by
 * RSA_public_encrypt */
static void init_random_generator ()
{
  /* using time as seed is very weak, but use if nothing else is available */
  unsigned int random_seed = time (NULL);
  int fd = open ("/dev/urandom", O_RDONLY);
  if (fd >= 0) {
    int bytes = read (fd, &random_seed, sizeof (random_seed));
    if (bytes != sizeof (random_seed))
      random_seed = time (NULL);
    close (fd);
  }
  srandom (random_seed);
}

/* fills in cypher with the signed and encrypted text.  key_id is the
 * ID identifying the recipient.  Also fills in the first ksize
 * bytes of the key, padding with zeros if ksize > key length.
 * returns the number of bytes in the cyphertext if successful -- if
 * the value returned > csize, the result does not fit and is not
 * stored in cypher.
 * returns -1 in case of failure.
 */
int sign_encrypt (char * text, int tsize, char * cypher, int csize,
                  int key_id, char * key, int ksize)
{
  if ((key_id < 0) || (key_id >= saved_keys)) {
    printf ("invalid key id %d, valid range 0..%d\n", key_id, saved_keys - 1);
    return -1;
  }
  RSA * rsa = keys [key_id].pubkey;

  if (num_private_keys < 1) {
    printf ("unable to sign message of size %d, no private keys\n", tsize);
    return -1;
  }
  RSA * signing_rsa = keys [0].privkey;

  if (tsize >= RSA_size (rsa) - 11) {
    /* fix later to use session keys.  This is OK for chat I think */
    printf ("error, message size %d larger than maximum allowed %d\n",
            tsize, RSA_size (rsa) - 11);
    return -1;
  }
/* it's 501 for 4096-bit = 512-byte keys
  printf ("maximum message size %d\n", RSA_size (rsa) - 11); */

  int total_bytes = RSA_size (rsa) + RSA_size (signing_rsa);
  if (csize < RSA_size (rsa) + RSA_size (signing_rsa)) {
    printf ("cyphertext size provided %d, but need %d + %d = %d bytes\n",
            csize, RSA_size (rsa), RSA_size (signing_rsa), total_bytes);
    return -1;
  }

  /* signature follows cyphertext */
  unsigned char * signature = cypher + RSA_size (rsa);
  unsigned int siglen;
  if (RSA_sign (NID_md5, text, tsize, signature, &siglen, signing_rsa) != 1) {
    unsigned long e = ERR_get_error ();
    printf ("RSA signature failed %ld/%s\n", e, ERR_error_string (e, NULL));
    return -1;
  }

  init_random_generator ();
  int bytes = RSA_public_encrypt (tsize, text, cypher, rsa, RSA_PKCS1_PADDING);
  if (bytes != RSA_size (rsa)) {
    unsigned long e = ERR_get_error ();
    printf ("RSA encryption failed %ld/%s %d\n", e, ERR_error_string (e, NULL),
            bytes);
    return -1;
  }
  return total_bytes;
}

/* returns 1 for a possible match, 0 for guaranteed no match */
static int matching_key (RSA * key, char * key_hint, int ksize)
{
  if ((ksize == 0) || (key_hint == NULL))
    return 1;
  int bn_size = BN_num_bytes (key->n);
  if (bn_size < ksize)
    return 0;
  int size = BN_num_bytes (key->n);
  char * buffer = malloc (size);
  if (buffer == NULL) {
    printf ("matching_key unable to allocate bignum buffer\n");
    return 1;
  }
  BN_bn2bin (key->n, buffer);
  int bytes = size;
  if (bytes > ksize)
    bytes = ksize;
  int comparison = memcmp (buffer, key_hint, bytes);
  free (buffer);
  if (comparison == 0)
    return 1;   /* match */
  return 0;  /* no match */
}

/* if successfully decrypts and verifies using the given key, fills in
 * text (up to tsize) and returns the total number of characters in
 * text (which may exceed tsize).
 *
 * if unable to verify, returns -1
 */
static int do_decrypt_verify (char * cypher, int csize, char * text, int tsize,
                              int key)
{
  if ((key < 0) || (key >= saved_keys)) {
    printf ("invalid key id %d, valid range 0..%d\n", key, saved_keys - 1);
    return -1;
  }
  RSA * signing_rsa = keys [key].pubkey;
  int sign_size = RSA_size (signing_rsa);

  if (num_private_keys < 1) {
    printf ("unable to decrypt message of size %d, no private keys\n", csize);
    return -1;
  }

  int i;
  /* try all the private keys for decryption */
  for (i = 0; i < num_private_keys; i++) {
    RSA * rsa = keys [i].privkey;
    if (RSA_size (rsa) + sign_size != csize)
      continue;   /* cannot be a match, try the next private key */
    char * sigp = cypher + RSA_size (rsa);
    char clear [2048 * 4];
    if (sizeof (clear) < RSA_size (rsa)) {
      printf ("RSA size %d too large for buffer size %d\n", RSA_size (rsa),
              sizeof (clear));
      return -1;
    }
    int decrypted = RSA_private_decrypt (RSA_size (rsa), cypher, clear,
                                         rsa, RSA_PKCS1_PADDING);
    if (decrypted <= 0)    /* unable to decrypt */
      continue;            /* try the next private key, if any */
    /* decryption successful, verify the signature */
    if (RSA_verify (NID_md5, clear, decrypted, sigp, sign_size, signing_rsa)) {
      int bytes = decrypted;
      if (decrypted > tsize)
        bytes = tsize;
      memcpy (text, clear, bytes);
      return decrypted;
    }
  }
  /* none of the keys verified, return */
  return -1;
}

/* if successfully decrypts and verifies, fills in text (up to tsize) and
 * returns the total number of characters in text (which may exceed tsize).
 * key_hint, if not NULL, is used to restrict the number of personal keys
 * which need to be tried.
 *
 * if unable to find a key to decrypt and verify, returns -1
 */
int decrypt_verify (char * cypher, int csize, char * text, int tsize,
                    char * key_hint, int ksize, int * key_id)
{
  int i;
  for (i = 0; i < saved_keys; i++) {
    if (matching_key (keys [i].pubkey, key_hint, ksize)) {
      int result = do_decrypt_verify (cypher, csize, text, tsize, i);
      if (result > 0) {
        if (key_id != NULL)
          *key_id = i;
        return result;
      }
    }
  }
  return -1;
}

/* verifies that the signature matches the text for the given public key.
 * the entire key must be provided.  This is intended for key exchange,
 * where in one phase both sides exchange the signatures of a string known
 * to both, then in the second phase, they send the (hopefully) matching
 * public key.
 *
 * returns 1 if successful, 0 if not matched, -1 for errors.
 */
int verify_signature (char * sig, int ssize, char * text, int tsize,
                      char * key, int ksize)
{
  return -1;
}

