/* creates, parses, and prints hello messages. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include "types.h"
#include "messages.h"

int header_size (struct allnet_route * route)
{
  int offset = 8 + 2 + 1;  /* total length, then type, then num_addrs */
  if (b1 (route->num_addrs) == 0)
    return offset;   /* no addresses, no "used" field */

  offset++;          /* "used" field */
  int i;
  for (i = 0; i < b1 (route->num_addrs); i++) {
    char buf [1];
    int size = store_addr (route->addrs + i, buf, 0);
    if (size >= 0) {  /* should be negative, since we did not give a buffer */
      printf ("error: unknown address size %d\n", size);
      exit (1);
    }
    /* printf ("address %d has size %d bytes\n", i, -size); */
    offset += (- size);
  }
  return offset;
}

static void store_header (char * buffer, int bsize,
                          allnet_data_len total_length, int type,
                          struct allnet_route * route, int hsize)
{
  if (bsize < hsize) {
    if (bsize != 0)
      printf ("unable to store %d header bytes in %d buffer bytes\n",
              hsize, bsize);
    return;  /* not much use continuing */
  }
  if (bsize >= 8)
    bs8 (buffer, total_length);
  if (bsize >= 10)
    bs2 (buffer + 8, type);
  int chars = store_route (route, buffer + 10, bsize - 10);
  if (chars <= 0) {
    printf ("error in store_header: %d header, %d buffer bytes, %d result\n",
            hsize, bsize, chars);
    exit (1);
  }
}

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_hello.  Store_hello also stores the allnet header. */

int store_hello (struct allnet_peer * myself,
              /* if num_peers is zero, peers may be NULL */
                 int num_peers, struct allnet_peer * peers,
                 struct allnet_route * route, char * buffer, int bsize)
{
  allnet_data_len total_length = 0;
  int hsize = header_size (route);
  int offset = hsize;
  int badresult = (offset >= bsize);
  int result;
  int use_offset = (badresult) ? 0 : offset;

  result = store_peer (myself, buffer + use_offset, bsize - use_offset);
  if (result == 0) return 0;
  if (result < 0) { badresult = 1; result = (-result); }
  offset += result;
  use_offset = (badresult) ? 0 : offset;

  if (offset + 2 <= bsize)
    bs2 (buffer + offset, num_peers);
  offset += 2;
  use_offset = (badresult) ? 0 : offset;

  if (num_peers > 0) {
    int i;
    for (i = 0; i < num_peers; i++) {
      result = store_peer (peers + i, buffer + use_offset, bsize - use_offset);
      if (result == 0) return 0;
      if (result < 0) { badresult = 1; result = (-result); }
      offset += result;
      use_offset = (badresult) ? 0 : offset;
    }
  }
  total_length = offset;
  printf ("store_hello, total length %lld = %d header + %lld data\n",
          total_length, hsize, total_length - hsize);

  if (! badresult)
    store_header (buffer, bsize, total_length, ALLNET_TYPE_HELLO, route, hsize);
  return total_length;
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the header is stored in *result  When called,
 * result->route.addrs should be NULL.  In any case, it will be malloc'd as
 * needed.
 */
int parse_header (char * buffer, int bsize, struct allnet_header * result)
{
  if (bsize < 11) return -1;
  result->total_length = b8 (buffer);
  result->allnet_type = b2 (buffer + 8);
  int num_addrs = b1 (buffer + 10);
  result->route.num_addrs [0] = num_addrs;
  if (num_addrs == 0) {
    result->route.used [0] = 0;
    result->route.addrs = NULL;
    return 11;
  }
  if (bsize < 12) return -1;
  result->route.addrs = malloc (sizeof (struct allnet_addr) * num_addrs);
  if (result->route.addrs == NULL) return -1;
  result->route.used [0] = b1 (buffer + 11);
  int offset = 12;
  int i;
  for (i = 0; i < num_addrs; i++) {
    int p = parse_addr (buffer + offset, bsize - offset,
                        result->route.addrs + i);
    if (p < 0) return -1;
    offset += p;
  }
  return offset;
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the message is stored in *result;
 */
int parse_hello (char * buffer, int bsize,
                 struct allnet_hello * result, int max_peers)
{
  int count = parse_peer (buffer, bsize, &(result->sender));
  if (count < 0) return -1;
  int offset = count;

  if (offset + 2 <= bsize) {
    result->num_peers = b2 (buffer + offset);
    offset += 2;
  } else {
    return -1;
  }

  if (result->num_peers > max_peers) {
    printf ("parsing hello message with %d peers, but max_peer is %d\n",
            result->num_peers, max_peers);
    return -1;
  }

  int i;
  for (i = 0; i < result->num_peers; i++) {
    count = parse_peer (buffer + offset, bsize - offset, result->peers + i);
    if (count < 0) return -1;
    offset += count;
  }

  return offset;
}

int hello_to_string (struct allnet_hello * hello,
                     char * string, int ssize)
{
  strncpy (string, "not implemented", ssize);
}

void print_hello (struct allnet_hello * hello, int eol)
{
  print_peer (&(hello->sender), eol);
  if (eol < 1) printf (", ");
  printf ("%d peers: ", hello->num_peers);
  if (eol >= 1) printf ("\n");
  if (hello->num_peers > 0) {
    int i;
    for (i = 0; i < hello->num_peers; i++) {
      print_peer (hello->peers + i, eol);
      if ((eol < 1) && (i + 1 < hello->num_peers)) printf (", ");
    }
  }
}


/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_data.  Store_data also stores the allnet header. */
int store_data (char * data, allnet_data_len length, int clear,
                struct allnet_route * route, char * buffer, int bsize)
{
  int hsize = header_size (route);
  int type = ALLNET_TYPE_DATA;
  if (clear)
    type = ALLNET_TYPE_CLEAR;
  allnet_data_len total_length = hsize + length;

  store_header (buffer, bsize, total_length, type, route, hsize);
  if (total_length <= bsize) {
    memcpy (buffer + hsize, data, length);
    return total_length;
  }
  return -total_length;
}

int data_to_string (allnet_data data, allnet_data_len len,
                    char * string, int ssize)
{
}

void print_data_message (allnet_data data, allnet_data_len len, int eol)
{
  printf ("%lld bytes of data:", len);
  if (eol >= 1)
    printf ("\n");
  else
    printf (" ");
  int i;
  for (i = 0; i < len; i++) {
    printf ("%02x", (data [i] & 0xff));
    if (i + 1 < len) printf (", ");
  }
  if (eol >= 1)
    printf ("\n");
}

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_signed.  Store_signed also stores the allnet header. */

int store_signed (char * data, allnet_data_len length,
                  int sig_type,
                  char * sig, allnet_data_len sig_length,
                  struct allnet_route * route, char * buffer, int bsize)
{
  allnet_data_len total_length = 0;
  int hsize = header_size (route);
  store_header (buffer, bsize, total_length, ALLNET_TYPE_SIGNED, route, hsize);
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, pointers to the part of the buffer containing the data
 * and signature, and their lengths, are stored in the last 4 parameters.
 */
int parse_signed (char * buffer, int bsize, int * sig_type,
                  allnet_data_len * data_len, char ** data,
                  allnet_data_len * sig_len, char ** sig)
{
  if (bsize < 8)
    return -1;
  *data_len = b8 (buffer);
  int offset = 8;

  if (bsize < offset + *data_len)
    return -1;
  *data = buffer + offset;
  offset += *data_len;

  if (bsize < offset + 4)
    return -1;
  *sig_type = b4 (buffer + offset);
  offset += 4;

  if (bsize < offset + 8)
    return -1;
  *sig_len = b8 (buffer + offset);
  offset += 8;
  
  if (bsize < offset + *sig_len)
    return -1;
  *sig = buffer + offset;
  offset += *sig_len;

  return offset;
}

int signed_to_string (struct allnet_signed * s, char * string, int ssize)
{
}

void print_signed (struct allnet_signed * s, int eol)
{
  print_data_message (s->message, s->msg_len, eol);
  printf ("signature type %d, ", s->sig_type);
  print_data_message (s->signature, s->sig_len, eol);
}



