/* messages.h: format for messages sent on the allnet network */

#ifndef ALLNET_MESSAGES_H
#define ALLNET_MESSAGES_H

#include "types.h"

#define ALLNET_PORT 	     (htons (0xa119))  /* ALLNet */
#define ALLNET_LOCAL_PORT    (htons (0xa11e))  /* ALLnEt */
#define ALLNET_WIFI_PROTOCOL (htons (0xa119))  /* ALLNet */

#define ALLNET_TYPE_NONE		0x00  /* no message, only header */
#define ALLNET_TYPE_HELLO		0x01
#define ALLNET_TYPE_DATA		0x02
#define ALLNET_TYPE_CLEAR		0x03
#define ALLNET_TYPE_SIGNED		0x04

#define MAX_PRIORITY			1000  /* 1000 is highest priority */
#define LOCAL_MSG_PRIORITY		900   /* near the top */
#define HELLO_PRIORITY			100   /* fairly low priority */
#define MIN_PRIORITY			0

/* don't send hello packets to a given peer any more than once every
 * 30 minutes */
#define HELLO_INTERVAL			(30 * 60)  /* at least 30 minutes */
/* #define HELLO_INTERVAL			100  /* for testing */

/* all allnet messages begin with an 8 byte total length, which counts
 * everything, including the header.  Next is a two byte allnet_type,
 * followed by a struct allnet_route.
 * Header and total length is always 12 bytes or more.
 */

struct allnet_header {
  allnet_data_len total_length;
  short allnet_type;
  struct allnet_route route;
};
/* given a route, how many bytes does the header need? */
extern int header_size (struct allnet_route * route);

/* what follows are the definitions of the message bodies.
 */

/* a hello message is normally sent on a directly connected wireless link,
 * or on an IP network.
 * On an IP network it will have an IP sender address, and a 2-hop route
 * which includes the sender and receiver.
 * On a directly connected wireless link it will have a key or broadcast
 * address as the sender.  The route may start out as an n-hop broadcast
 * address, and be built up into a return route including a number of
 * key addresses, one per hop (perhaps with some intermediate broadcast
 * segments).
 * Either kind of message may have any number of peers. */
/* a recipient of a hello message may or may not reply with the same
 * kind of message, but in any case should keep track of senders and not send
 * a hello message to the same peer more than once every half hour or so,
 * i.e. HELLO_INTERVAL.  For a simple implementation, it can just send
 * hello messages once every half hour, or it can be more specific.
 */
struct allnet_hello {
  /* my address and services I want to tell you about */
  struct allnet_peer sender;
  short num_peers;
  /* zero or more peers I want to tell you about */
  struct allnet_peer peers [0];
};

/* a data message with encrypted contents.  Hopefully, the recipient
 * knows what to do with it.  The length is taken from the header,
 * and is usually greater than 0.
 */
typedef char allnet_data [0];

/* a message with clear contents.  These should generally be
 * forwarded to authorities.  If shown to the user, the user
 * should be given some training regarding the likelihood that
 * this message may be misleading or untruthful.
 *
 * note the actual structure is identical to allnet_data.
 */

/* a message from someone claiming some externally visible ID. */
struct allnet_signed {
  allnet_data_len msg_len;
  allnet_data message;          /* in the clear */
  int sig_type;                 /* not yet defined */
  allnet_data_len sig_len;
  allnet_data signature;        /* sig and certificate and anything else */
};

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the header is stored in *result  When called,
 * result->route.addrs should be NULL.  In any case, it will be malloc'd as
 * needed.
 */
extern int parse_header (char * buffer, int bsize,
                         struct allnet_header * result);

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_hello.  Store_hello also stores the allnet header. */

extern int store_hello (struct allnet_peer * myself,
                     /* if num_peers is zero, peers may be NULL */
                        int num_peers, struct allnet_peer * peers,
                        struct allnet_route * route, char * buffer, int bsize);
extern int parse_hello (char * buffer, int bsize,
                        struct allnet_hello * result, int max_peers);
extern int hello_to_string (struct allnet_hello * hello,
                            char * string, int ssize);
extern void print_hello (struct allnet_hello * hello, int eol);

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_data.  Store_data also stores the allnet header. */

extern int store_data (char * data, allnet_data_len length, int clear,
                       struct allnet_route * route, char * buffer, int bsize);
extern int data_to_string (allnet_data data, allnet_data_len len,
                           char * string, int ssize);
extern void print_data_message (allnet_data data, allnet_data_len len, int eol);

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_signed.  Store_signed also stores the allnet header. */

extern int store_signed (char * data, allnet_data_len length,
                         int sig_type,
                         char * sig, allnet_data_len sig_length,
                         struct allnet_route * route, char * buffer, int bsize);
/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, pointers to the part of the buffer containing the data
 * and signature, and their lengths, are stored in the last 4 parameters.
 */
extern int parse_signed (char * buffer, int bsize, int * sig_type,
                         allnet_data_len * data_len, char ** data,
                         allnet_data_len * sig_len, char ** sig);
extern int signed_to_string (struct allnet_signed * s,
                             char * string, int ssize);
extern void print_signed (struct allnet_signed * s, int eol);

#endif /* ALLNET_MESSAGES_H */
