/* synchronized queue */
/* this queue synchronizes all accesses (sort of like a Hoare monitor)
 * so callers can perform calls in any order, and the messages are
 * correctly added to or removed from the queue.
 *
 * the queue is organized by priority, so elements with higher priority
 * are returned before elements with lower priority.  Objects with the
 * same priority are returned in the order of insertion.
 *
 * Objects are kept with their size as well as the time of insertion.
 * Objects may be removed from the queue either explicitly, or from
 * time to time by calling cleanup, which removes elements older than
 * a certain time and taking up too much space.
 *
 * it is assumed that elements are malloc'd, and any elements removed
 * from the queue (and not returned to the caller) will be free'd
 */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>

#include "synqueue.h"

struct syn_queue_element {
  void * value;
  int size;
  int priority;
  struct timeval time;
};

struct syn_queue {
  struct syn_queue_element * queue;
  int num_slots;
  int num_elements;
  int total_size;
  int max_size;
  pthread_mutex_t lock;
  pthread_cond_t condition;
};

/* #define DEBUG_LOCKS */

#ifdef DEBUG_LOCKS
static void print_releasing (char * fn_name)
{
  printf ("function %s releasing lock\n", fn_name);
}
static void print_locking (char * fn_name)
{
  printf ("function %s requesting lock\n", fn_name);
}
static void print_locked (char * fn_name)
{
  printf ("function %s has acquired lock\n", fn_name);
}

#define get_lock(l, n) { print_locking (n); pthread_mutex_lock (&l); print_locked (n); }
#define release_lock_return(l, v, n) { print_releasing (n); pthread_mutex_unlock (&l); return v; }
#define release_lock(l, n) { print_releasing (n); pthread_mutex_unlock (&l); }
#else  /* not debugging */
#define get_lock(l, n) { pthread_mutex_lock (&l); }
#define release_lock_return(l, v, n) { pthread_mutex_unlock (&l); return v; }
#define release_lock(l, n) { pthread_mutex_unlock (&l); }
#endif  /* debugging */

/* if optional_max_size is not zero, it is the maximum size, in bytes,
 * of all the elements in the queue.  Adding any element beyond the maximum
 * size results in the automatic removal of the lowest priority element(s),
 * until the maximum size is honored.  Objects larger than the maximum size
 * cannot be added to the queue.
 *
 * new_syn_queue returns NULL in case of error.
 */
struct syn_queue * new_syn_queue (int optional_max_size)
{
  struct syn_queue * result = malloc (sizeof (struct syn_queue));
  if (result != NULL) {
    result->max_size = optional_max_size;
    result->total_size = 0;
    result->num_elements = 0;
    int slots = 10;  /* to start with */
    result->num_slots = slots;
    result->queue = malloc (sizeof (struct syn_queue_element) * slots);
    if (result->queue == NULL) {
      free (result);
      result = NULL;
    }
    pthread_mutex_init (&(result->lock), NULL);
    pthread_cond_init (&(result->condition), NULL);
  }
  return result;
}

void free_syn_queue (struct syn_queue * q)
{
  pthread_mutex_destroy (&(q->lock));
  if (q->queue != NULL) {
    int i;
    for (i = 0; i < q->num_slots; i++)
      if (q->queue [i].value != NULL)
        free (q->queue [i].value);
    free (q->queue);
  }
  free (q);
}

/* returns 1 if the queue is empty, 0 otherwise */
int queue_empty (struct syn_queue * q)
{
  /* atomic, single access, no need for mutex */
  return (q->num_elements == 0);
}


/* returns the number of bytes in elements stored in the queue */
int queue_size (struct syn_queue * q)
{
  /* atomic, single access, no need for mutex */
  return q->total_size;
}


/* returns the number of elements stored in the queue */
int queue_elements (struct syn_queue * q)
{
  /* atomic, single access, no need for mutex */
  return q->num_elements;
}


/* returns 0 if the queue is empty, or the priority otherwise (note that
 * the highest priority could be zero).
 * If it does not return zero, and for each if it is not NULL, also fills in
 * the element value, size, and priority
 */
int queue_priority (struct syn_queue * q,
                    void ** value, int * size, int * priority)
{
  get_lock (q->lock, "queue_priority");
  int result = 0;
  if (! queue_empty (q)) {
    result = q->queue [0].priority;
    if (value != NULL) *value = q->queue [0].value;
    if (size != NULL) *size = q->queue [0].size;
    if (priority != NULL) *priority = q->queue [0].priority;
  }
  release_lock_return (q->lock, result, "queue_priority");
}

/* returns 0 if the times are equal, 1 if t1 > t2 (t1 after t2), and
 * -1 if t1 < t2 (t1 before t2)
 */
static int compare_timeval (struct timeval * t1, struct timeval * t2)
{
  if (t1->tv_sec  > t2->tv_sec)
    return 1;
  if (t1->tv_sec  < t2->tv_sec)
    return -1;
  if (t1->tv_usec > t2->tv_usec)
    return 1;
  if (t1->tv_usec < t2->tv_usec)
    return -1;
  return 0;
}

/* sets *time the earliest time of any element stored in the queue, unless
 * the queue is empty or time is NULL.  If the queue is not empty,
 * and for each if it is not NULL, also fills in
 * the element value, size, and priority
 */
void queue_earliest (struct syn_queue * q, struct timeval * time,
                     void ** value, int * size, int * priority)
{
  get_lock (q->lock, "queue_earliest");
  if (q->num_elements > 0) {
    int index = 0;
    int i;
    for (i = 1; i < q->num_elements; i++)
      if (compare_timeval (&(q->queue [index].time), &(q->queue [i].time)) > 0)
        /* i is earlier than index */
        index = i;
    if (time != NULL) *time = q->queue [index].time;
    if (value != NULL) *value = q->queue [index].value;
    if (size != NULL) *size = q->queue [index].size;
    if (priority != NULL) *priority = q->queue [index].priority;
  }
  release_lock (q->lock, "queue_earliest");
}

/* sets *time the latest time of any element stored in the queue, unless
 * the queue is empty or time is NULL.  If the queue is not empty,
 * and for each if it is not NULL, also fills in
 * the element value, size, and priority
 */
void queue_latest (struct syn_queue * q, struct timeval * time,
                     void ** value, int * size, int * priority)
{
  get_lock (q->lock, "queue_latest");
  if (q->num_elements > 0) {
    int index = 0;
    int i;
    for (i = 1; i < q->num_elements; i++)
      if (compare_timeval (&(q->queue [index].time), &(q->queue [i].time)) < 0)
        /* i is later than index */
        index = i;
    if (time != NULL) *time = q->queue [index].time;
    if (value != NULL) *value = q->queue [index].value;
    if (size != NULL) *size = q->queue [index].size;
    if (priority != NULL) *priority = q->queue [index].priority;
  }
  release_lock (q->lock, "queue_latest");
}

/* index should be >= 0 and < q->num_elements.
 * The caller should hold the lock */
static void remove_at_index (struct syn_queue * q, int index, int free_value)
{
  if ((q->num_elements <= 0) || (index < 0) || (index > q->num_elements)) {
    printf ("remove_at_index (%d), but queue has %d elements\n",
            index, q->num_elements);
    return;
  }
  if (free_value)
    free (q->queue [index].value);
  q->total_size -= q->queue [index].size;
  int i;
  for (i = index + 1; i < q->num_elements; i++)
    q->queue [i - 1] = q->queue [i];
  q->num_elements--;
}

static void add_element_at_index (struct syn_queue * q, void * value, int size,
                                  int priority, int index)
{
  if (q->num_elements == q->num_slots) {   /* must expand */
    int slots = 10 + q->num_slots * 10;
    q->queue = realloc (q->queue, sizeof (struct syn_queue_element) * slots);
    if (q->queue == NULL) {
      printf ("realloc of queue failed for %d slots\n", slots);
      exit (1);
    }
    q->num_slots = slots;
  }
  int i;
  for (i = q->num_elements; i > index; i--)
    q->queue [i] = q->queue [i - 1];
  q->queue [index].value = value;
  q->queue [index].size = size;
  q->queue [index].priority = priority;
  gettimeofday (&(q->queue [index].time), NULL);
  q->total_size += size;
  q->num_elements++;
/*
  printf ("added element at index %d, size is now %d with %d elements\n",
          index, q->total_size, q->num_elements);
*/
}

/* not used */
static void remove_oldest (struct syn_queue * q)
{
  if (q->num_elements == 0)
    return;

  struct timeval earliest;
  int earliest_index;
  int i;

  earliest_index = 0;
  earliest = q->queue [0].time;
  for (i = 1; i < q->num_elements; i++) {
    if (compare_timeval (&earliest, &(q->queue [i].time)) > 0) {
       /* new is earlier */
      earliest = q->queue [i].time;
      earliest_index = i;
    }
  }
  struct timeval now;
  gettimeofday (&now, NULL);
  int delta = (now.tv_sec - earliest.tv_sec) * 1000000 +
              (now.tv_usec - earliest.tv_usec);
  printf ("removing oldest element inserted %d.%06d seconds ago\n",
          delta / 1000000, delta % 1000000);
  remove_at_index (q, earliest_index, 1);
  printf ("remaining size is %d\n", q->total_size);
}

/* returns 1 if removed, 0 if queue is empty or min_priority < priority
 * of any element in the queue
 */
static int remove_lowest_priority (struct syn_queue * q, int min_priority)
{
  if (q->num_elements == 0)
    return 0;

  /* lowest priority element should be at the end */
  int i = q->num_elements - 1;
  int lowest_priority = q->queue [i].priority;
  if (lowest_priority > min_priority)
    return 0;  /* priority is higher than what we want to insert */
  do {
    i--;
  } while ((i >= 0) && (q->queue [i].priority == lowest_priority));
  /* i + 1 is the index of the oldest element with the lowest priority */
  int lowest_index = i + 1;

  struct timeval now;
  gettimeofday (&now, NULL);
  int delta = (now.tv_sec - q->queue [lowest_index].time.tv_sec) * 1000000 +
              (now.tv_usec - q->queue [lowest_index].time.tv_usec);
  printf ("removing priority %d element inserted %d.%06d seconds ago\n",
          lowest_priority, delta / 1000000, delta % 1000000);
  remove_at_index (q, lowest_index, 1);
  printf ("remaining size is %d\n", q->total_size);
  return 1;
}

/* size should be > 0, priority >= 0 */
/* returns the size in case of success,
 * -1 in case the element could not be added
 */
int queue_add (struct syn_queue * q, void * element, int size, int priority)
{
  if ((size <= 0) || (priority < 0)) {
    printf ("size %d (should be > 0), priority %d (should be >= 0)\n",
            size, priority);
    return -1;
  }

  get_lock (q->lock, "queue_add");
  if ((q->max_size != 0) && (size > q->max_size)) {
    printf ("size %d, but queue max size is %d\n", size, q->max_size);
    release_lock_return (q->lock, -1, "queue_add");
  }
  if (q->max_size != 0) {
    while ((q->total_size + size > q->max_size) &&
           (remove_lowest_priority (q, priority)))
      ;
    if (q->total_size + size > q->max_size)
 /* nothing of lower/same priority in the queue, and no room to insert it */
      return -1;
  }
  int i;
  int added = 0;
  for (i = 0; i < q->num_elements; i++) {
    if (priority > q->queue [i].priority) { /* insert here */
      add_element_at_index (q, element, size, priority, i);
      added = 1;
      break;
    }
  }
  if (! added)
    add_element_at_index (q, element, size, priority, q->num_elements);
  pthread_cond_signal (&(q->condition));  /* wake up anyone waiting */
  release_lock_return (q->lock, size, "queue_add");
}

/* called with locks held */
static void * get_nth (struct syn_queue * q, int index,
                       int * size, int * priority)
{
  void * result = NULL;
  if (q->num_elements > index) {
    if (size != NULL) *size = q->queue [index].size;
    if (priority != NULL) *priority = q->queue [index].priority;
    result = q->queue [index].value;
  }
  return result;
}


/* returns (and removes) the first (highest priority) element in the queue,
 * or NULL if the queue is empty */
void * queue_get (struct syn_queue * q, int * size, int * priority)
{
  get_lock (q->lock, "queue_get");
  void * result = get_nth (q, 0, size, priority);
  if (result != NULL)
    remove_at_index (q, 0, 0);
  release_lock_return (q->lock, result, "queue_get");
}

static void abort_on_error (int error)
{
  if (error != 0) {
    printf ("error in queue_wait: pthread_cond_wait returned %d", error);
    printf ("/%s\n", strerror (error));
    exit (1);
  }
}

/* same as queue_get, but if there is nothing in the queue, waits until
 * something is available, so never returns NULL. */
void * queue_wait (struct syn_queue * q, int * size, int * priority)
{
  get_lock (q->lock, "queue_wait");
  void * result = get_nth (q, 0, size, priority);
  while (result == NULL) {
    abort_on_error (pthread_cond_wait (&(q->condition), &(q->lock)));
    result = get_nth (q, 0, size, priority);
  }
  remove_at_index (q, 0, 0);
  release_lock_return (q->lock, result, "queue_wait");
}

/* same as queue_get, but does not remove. */
void * queue_first (struct syn_queue * q, int * size, int * priority)
{
  get_lock (q->lock, "queue_first");
  void * result = get_nth (q, 0, size, priority);
  release_lock_return (q->lock, result, "queue_first");
}

#if 0  /* not used, I think */
/* same as queue_first, but nth element (returning NULL if no such element). */
/* use index 0 for the first element, 1 for the second, etc.
 * With index == 0, equivalent to queue_first */
void * queue_nth (struct syn_queue * q, int index, int * size, int * priority)
{
  get_lock (q->lock, "queue_nth");
  void * result = get_nth (q, index, size, priority);
  release_lock_return (q->lock, result, "queue_nth");
}
#endif /* not used, I think */

/* same as queue_first, but returns a malloc'd array of the first n values.
 * returns a value 0 <= i <= n giving the number of elements returned.
/* use n = 1 for one element, 2 for the two elements, etc. */
/* the sizes and priorities arrays (unless sizes/priorities is/are NULL)
 * are also malloc'd, should be freed when no longer needed.
 */
int queue_n (struct syn_queue * q, int n,
             void * * * elements, int ** sizes, int ** priorities)
{
  get_lock (q->lock, "queue_n");
  if (q->num_elements < n) n = q->num_elements;
  void * * my_e = NULL;
  int * my_s = NULL;
  int * my_p = NULL;
  if (elements != NULL) {
    my_e = malloc (n * sizeof (void *));
    if (my_e == NULL) {
      printf ("queue_n unable to allocate elements array %d\n", n);
      release_lock_return (q->lock, 0, "queue_n");
    }
  }
  if (sizes != NULL) {
    my_s = malloc (n * sizeof (int));
    if (my_s == NULL) {
      printf ("queue_n unable to allocate sizes array %d\n", n);
      if (my_e != NULL) free (my_e);
      release_lock_return (q->lock, 0, "queue_n/2");
    }
  }
  if (priorities != NULL) {
    my_p = malloc (n * sizeof (int));
    if (my_p == NULL) {
      printf ("queue_n unable to allocate priorities array %d\n", n);
      if (my_e != NULL) free (my_e);
      if (my_s != NULL) free (my_s);
      release_lock_return (q->lock, 0, "queue_n/3");
    }
  }

  /* all that is needed, allocated */
  int index;
  for (index = 0; index < n; index++) {
    if (my_e != NULL) my_e [index] = q->queue [index].value;
    if (my_s != NULL) my_s [index] = q->queue [index].size;
    if (my_p != NULL) my_p [index] = q->queue [index].priority;
  }

  if (elements   != NULL) *elements   = my_e;
  if (sizes      != NULL) *sizes      = my_s;
  if (priorities != NULL) *priorities = my_p;
  release_lock_return (q->lock, n, "queue_n/4");
}

/* applies the given function to the given argument, and every element
 * of the queue */
void queue_map (struct syn_queue * q, void * arg,
                void (* fun) (void * v1, void * v2))
{
  get_lock (q->lock, "queue_map");
  int i;
  for (i = 0; i < q->num_elements; i++)
    fun (arg, q->queue [i].value);
  release_lock (q->lock, "queue_map");
}

/* remove all elements for which the given function returns 1 when
 * applied to the given argument and a queue element. 
 * Return the number of elements removed.  Removed element are freed */
int queue_remove (struct syn_queue * q, void * arg,
                  int (* match) (void * v1, void * v2))
{
  get_lock (q->lock, "queue_remove");
  int result = 0;
  int i = 0;
  while (i < q->num_elements) {
    if (match (arg, q->queue [i].value) == 1) {
      remove_at_index (q, i, 1);
      result++;
    } else {
      i++;
    }
  }
  release_lock_return (q->lock, result, "queue_remove");
}

#ifdef TEST_QUEUE

#include <string.h>

static char * new_copy (const char * original)
{
  int size = strlen (original) + 1;
  char * result = malloc (size);
  if (result == NULL) {
    printf ("unable to allocate copy of '%s'\n", original);
    exit (1);
  }
  strcpy (result, original);
  return result;
}

static void print_queue (struct syn_queue * q)
{
  get_lock (q->lock, "print_queue");
  printf ("queue has %d bytes %d items\n", q->total_size, q->num_elements);
  int i;
  for (i = 0; i < q->num_elements; i++) {
    struct timeval now;
    gettimeofday (&now, NULL);
    int delta = (now.tv_sec - q->queue [i].time.tv_sec) * 1000000 +
                (now.tv_usec - q->queue [i].time.tv_usec);
    printf ("q [%d] has '%s' = %d bytes, priority %d, %d.%06d seconds ago\n",
            i, q->queue [i].value, q->queue [i].size, q->queue [i].priority, 
            delta / 1000000, delta % 1000000);
  }
  pthread_mutex_unlock (&(q->lock));
}

int main (int argc, char * argv)
{
  struct syn_queue * q = new_syn_queue (20);

  char content1 [] = "hello world";
  char * c1 = new_copy (content1);
  char content2 [] = "foo";
  char * c2 = new_copy (content2);
  char content3 [] = "bar";
  char * c3 = new_copy (content3);
  char content4 [] = "baz";
  char * c4 = new_copy (content4);

  print_queue (q);
  queue_add (q, c1, sizeof (content1), 100);
  print_queue (q);
  queue_add (q, c2, sizeof (content2), 1);
  print_queue (q);
  queue_add (q, c3, sizeof (content3), 5);
  print_queue (q);
  queue_add (q, c4, sizeof (content4), 3);
  print_queue (q);

  int s, p;
  char * r = (char *) (queue_get (q, &s, &p));
  printf ("first element is '%s' with size %d, priority %d\n", r, s, p);
  print_queue (q);

  r = (char *) (queue_get (q, &s, &p));
  printf ("second element is '%s' with size %d, priority %d\n", r, s, p);
  print_queue (q);

  r = (char *) (queue_get (q, &s, &p));
  printf ("last element is '%s' with size %d, priority %d\n", r, s, p);
  print_queue (q);
}

#endif /* TEST_QUEUE */

