/* types.h: format for data used in messages sent on the allnet network */

/*
* Copyright (c) 2012, Edoardo Biagioni
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef ALLNET_TYPES_H
#define ALLNET_TYPES_H

/* allnet_ types reflect the layout in the message, to the extent possible
 * in C (hard to express variable sizes in C).
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>

/* in every case, the hop count, if there is one, is the first byte */
/* 1 byte */
struct allnet_broadcast_addr {
  char hops [1];         /* one byte to tell us how many more hops */
};

/* 3-18 bytes.  The ID is selected at random, and might not be unique */
struct allnet_unicast_addr {
  char hops [1];         /* one byte to tell us how many more hops */
  char num_bytes [1];    /* one byte to tell us how many bytes (at most 16) */
  char id [16];          /* n bytes, 16 >= n > 0 */
};

/* 2 to 18 bytes */
struct allnet_pubkey_addr {
  char hops [1];         /* one byte to tell us how many more hops */
  char num_bits [1];     /* one byte to tell us how many bits (at most 128) */
  char pubkey [16];      /* n bytes, where n = (num_bits + 7) / 8 */
};

/* ports and IP addresses are always in network byte order */
/* 6 bytes */
struct allnet_ipv4_addr {
  struct in_addr ip4;    /* 4 bytes for the IP address */
  short port;            /* two bytes for the port number */
};

/* 22 bytes */
struct allnet_ipv6_addr {
  struct in6_addr ip6;   /* 16 bytes for the IP address */
  short port;            /* two bytes for the port number */
  int scope;             /* address scope -- may be needed */
};

/* coordinates are given as 32-bit signed integers, each representing
 * a fraction of +-180 degrees.  For latitude, this means half the space
 * is unused (there are only +- 90 degrees of latitude).
 * since the circumference of the earth is about 40,075,200m (at the equator),
 * this gives a resolution of about 40million m / 2^32, or about 1cm,
 * which is enough for practical purposes.
 * So for example, 0x20000000 is +1/4, meaning +45 degrees.
 *
 * In the unlikely event of needing to specify exactly
 * 180 degrees, use 0x80000000, which is -1 or exactly -180 degrees.
 *
 * WGS84 is used as the reference system.
 *
 * the minimum needed for a geographic address is two coordinates, lat and lon.
 * normally a radius will also be given, so three values is the usual.
 *
 * Occasionally a height, and then usually the range of heights,
 * might also be given.  Height and range are given in centimeters,
 * giving a maximum height of +-21,474km.  This is considerably
 * larger than the radius of the earth (for negative distances),
 * so it seems likely that the available range can be used for any useful
 * earth-based coordinates.
 *
 * sometimes it might be useful to give a height but not a radius.
 * in this case, the radius is set to 0x8000000000000000.
 */
/* 8, 12, 16, or 20 bytes */
struct allnet_geo_addr {
  int center_lat;   /* as a fraction of +-180 deg */
  int center_lon;   /* as a fraction of +-180 deg */
  int radius;       /* as a fraction of +-180 deg */
  int height;       /* in centimeters */
  int height_range; /* in centimeters */
};

/* 2 or more bytes */
struct allnet_addr {
  short discrim;         /* two bytes to discriminate among addrs */
#define ALLNET_NO_ADDR_ID		0x00
  union {
#define ALLNET_BCAST_ADDR_ID		0x01
    struct allnet_broadcast_addr bcast;
#define ALLNET_UNICAST_ADDR_ID		0x02
    struct allnet_unicast_addr ucast;
#define ALLNET_KEY_ADDR_ID		0x03
    struct allnet_pubkey_addr pubkey;
#define ALLNET_IP4_ADDR_ID		0x04
    struct allnet_ipv4_addr ipv4;
#define ALLNET_IP6_ADDR_ID		0x05
    struct allnet_ipv6_addr ipv6;
#define ALLNET_GEO_ADDR_ID_LL		0x10
#define ALLNET_GEO_ADDR_ID_LLR		0x11
#define ALLNET_GEO_ADDR_ID_LLRH		0x12
#define ALLNET_GEO_ADDR_ID_LLRHR	0x13
    struct allnet_geo_addr  geo;
  } u;
#define bc_addr		u.bcast
#define uc_addr		u.ucast
#define pk_addr		u.pubkey
#define i4_addr		u.ipv4
#define i6_addr		u.ipv6
#define geo_addr	u.geo
#define all_hops	u.bcast.hops
};

/* an allnet route is a form of loose source routing
 * each subsequent destination is expected or hoped to know how to
 * reach the following destination
 */
/* 1+ bytes -- num_addrs could be zero, in which case "used" is not included */
struct allnet_route {
  char num_addrs [1];     /* one byte to tell us how many addrs */
  char used [1];          /* one byte tells us how many addrs have been used */
  /* if we want to be able to construct a reverse path, then "used"
   * will be at least 1, and "num_addrs [0]" will be the address of the
   * original sender.
   * The packet will only be forwarded if "used" <= "num_addrs"
   */
  struct allnet_addr * addrs;
}; 

/* messages can be up to 2^64-1 bytes, though usually much less */
typedef unsigned long long int allnet_data_len;    /* 8 bytes, 64 bits */

/* list of services provided by our peers, perhaps indirectly */
typedef long long int allnet_services;    /* 8 bytes, 64 bits */

#define INTERNET_SERVICE	(1 << 0)
#define LOCATION_SERVICE	(1 << 1)
#define PHONE_MESSAGE_SERVICE	(1 << 2)
#define SEARCH_SERVICE		(1 << 3)

struct allnet_peer {
  allnet_services provides;
  struct allnet_addr peer;
};

/* useful for accessing multi-byte quantities */
#define b1(a)   ((a) [0] & 0xff)
#define b2(a)   ((b1 (a) << 8) | (b1 (a + 1)))
#define b4(a)   ((b2 (a) << 16) | (b2 (a + 2)))
#define b8(a)   ((((long long int) (b4 (a))) << 32) | \
                 ((long long int) (b4 (a + 4))))

#define bs1(a, v)   (*(a) = (v) & 0xff)
#define bs2(a, v)   { bs1 (a, ((v) >> 8)); bs1 ((a) + 1, (v)); }
#define bs4(a, v)   { bs2 (a, ((v) >> 16)); bs2 ((a) + 2, (v)); }
#define bs8(a, v)   { bs4 (a, ((v) >> 32)); bs4 ((a) + 4, (v)); }

/* the following four functions implemented in addr.c */

/* stores an address for sending on the net.  
 * if successful, returns the number of bytes used.
 * If bsize is insufficient, returns - the size needed,
 * e.g. -8 for an IPv4 address.
 * zero is returned in case of other errors.
 * in case of any error, the buffer may be garbled, do not use.
 */
extern int store_addr (struct allnet_addr * addr, char * buffer, int bsize);

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the address is stored in *result;
 */
extern int parse_addr (char * buffer, int bsize, struct allnet_addr * result);

/* makes the address printable, adding a '\0' at the end, returning the
 * number of bytes written (including the '\0').
 * If ssize is insufficient, returns - the size needed,
 * e.g. -21 for an IPv4 address such as 123.123.123.123/12345.
 * zero is returned in case of other errors.
 * in case of any error, the string may be garbled, do not use.
 */
extern int addr_to_string (struct allnet_addr * addr, char * string, int ssize);

/* prints the address, optionally with a newline at the end (if eof > 0) */
extern void print_addr (struct allnet_addr * addr, int eol);

/* the following four functions implemented in route.c.  The definitions
 * match the corresponding functions in addr.c */

extern int store_route (struct allnet_route * route, char * buffer, int bsize);
extern int parse_route (char * buffer, int bsize, struct allnet_route * result);
extern int route_to_string (struct allnet_route * route,
                            char * string, int ssize);
extern void print_route (struct allnet_route * route, int eol);

/* the following four functions implemented in peer.c.  The definitions
 * match the corresponding functions in addr.c */

extern int store_peer (struct allnet_peer * peer, char * buffer, int bsize);
extern int parse_peer (char * buffer, int bsize, struct allnet_peer * result);
extern int peer_to_string (struct allnet_peer * peer,
                           char * string, int ssize);
extern void print_peer (struct allnet_peer * peer, int eol);

/* the following functions are only for addrs, not for other types */
/* also implemented in addrs.c */

/* returns 1 if the addresses are the same, 0 if they differ */
extern int same_addr (struct allnet_addr * a1, struct allnet_addr * a2);

/* returns 1 if the address 1 matches address 2, 0 if it cannot. */
/* this is mostly the same as same_addr, but for keys and unicast
 * it allows a more generic a1 to match a more specific a2 */
extern int matches_addr (struct allnet_addr * a1, struct allnet_addr * a2);

#endif /* ALLNET_TYPES_H */

